#!/usr/bin/env python3
#
# Set-up ssh X11 forwarding for copy-paste support
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# # Requires Python package "pyperclip-1.8.2"
# xhost +
# ssh -Y username@hostname
# export DISPLAY=localhost:10.0
#
# Notes
# ~~~~~
# - Watch out for "stream.py:Frame()" dataclass and
#   "dashboard.py" use of "asciimatics.widgets.Frame()"
#
# Debugging approach #1
#   aiko.process.message.publish("$AIKO_NAMESPACE/DASHBOARD", f"Debug message")
#
#   mosquitto_sub -h $AIKO_MQTT_HOST -t $AIKO_NAMESPACE/DASHBOARD -v
#
# Debugging approach #2
#   from aiko_services.main.utilities import *
#   print_error("I've got a bad feeling about this !")
#
#   aiko_dashboard 1>/dev/null  # Debug messages only
#   aiko_dashboard 2>/dev/null  # TUI only
#
# Debugging approach #3
#   import logging
#   logging.basicConfig(format='%(message)s')
#   log = logging.getLogger(__name__)
#   log.warning("It's a trap !\r")
#
#   aiko_dashboard 1>/dev/null  # Debug messages only
#   aiko_dashboard 2>/dev/null  # TUI only
#
# Debugging approach #4
#   Write a custom Dashboard plugin for the Service / Actor
#
# To Do Elsewhere !
# ~~~~~~~~~~~~~~~~~
# * Turn Registrar into an ECProducer
# * Integrate into Ray HLActor ... Service, Actor, ECProducer, ECConsumer
#
# To Do
# ~~~~~
# * AI REPL --> LISP/Ollama --> Aiko Services analysis logs and MQTT messages
#
# * BUG: Dashboard doesn't display variables that contain whitespace !
#        Variable doesn't appear at all :(
#
# * BUG: When Dashboard exits, must clean-up (unshared) the ECConsumer !
#        Also, check that the ECProducer lease(s) expire as required
#
# * BUG: Dashboard isn't terminating ECConsumer lease extend timer :(
#
# - BUG: If currently selected Service terminates, then Dashboard doesn't
#        update selected Service's veriables section
#
# - BUG: Dashboard "Kill Service" command assumes running on same system
#        as the Service.  Should send "terminate" MQTT message to the
#        Service.  Or find a ProcessManager running on the correct host ?
#
# * FIX: Whenever DashboardFrame or LogFrame is destroyed and recreated
#        due to ResizeScreenError, all handlers need to be removed.
#        Provide DashboardFrame.cleanup() and LogFrame.cleanup(),
#        which are invoked by ResizeScreenError
#
# * FIX: Variable update dialogue replace with "maps.py:EnterLocation(Frame)"
#
# * FIX: Enable Service History selected "topic path" to clipboard
# * FIX: Enable Service History selected "topic path" can show LogFrame
# * FIX: Enable Service History multiple Service selection for logging
#
# * Write a file or template (for a script) for every Service / Actor seen
# *** "w" command for on-demand write file
# *** Provide a simple script that extracts the last topic_path and uses it
#
# - Show usage statistics: mosquitto messages, Service messages / uptime
#
# - LogFrame should support LogLevelPopupMenu to update selected Service
# - LogFrame display and allow the Service's "log_level" to be updated
#
# * If Registrar isn't available, then display "Waiting for Registrar"
# - Secondary Registrars should periodically send a non-retained message to
#     the TOPIC_REGISTRAR_BOOT topic ... (secondary found ...) for Dashboard
# - Dashboard should show if ...
#   - TOPIC_REGISTRAR_BOOT indicates that there is no primary Registrar
#   - TOPIC_REGISTRAR_BOOT indicates that there is a primary Registrar,
#       but the primary is not responding to (history ...) or (share ...)
#   - TOPIC_REGISTRAR_BOOT indicates there are secondary Registrar(s)
#       and show their details
#
# * Keyboard "H": Toggles between "Services History" and "Service Log"
#
# * Show Services running: time_add and time running
# * Show Services history: tags, time_add, time_remove and time running
#
# - ArchiveService should record "+/+/+/log" and removed Services ...
#   - Dashboard History section can ECConsumer the removed Services
#
# - Consider how to efficiently provide Service summary lifecycle states
#
# - Selecting (mouse or tab key) Service allows ...
#   - Toggle show/hide Services with specific field values (share * * * *)
#   - Subscribe to MQTT messages ("s" key) from topic (/#, /out, /state)
#   - Publish MQTT message ("p" key) to topic (/#, /in, /control, ...)
#
# - Service variable details should sort variable names alphabetically
# - Toggle show/hide of Service variables "services.*" visually redundant
# - Toggle show/hide Service variables with specific names (regex)
# - Allow Service variables to be added and removed
#
# - Dashboard Web browser (JavaScript) implementation using MQTT / WebSockets
#   - Service / ECProducer / ECConsumer JavaScript implementation
#   - Integrate Dashboard into HighLighter Web !
#
# Service variables that Services should have ...
# - lifecycle state: ...
# - log level: info, debug, ...
# - statistics: busy/idle time, mailbox queue size, message count, uptime
#
# Actors that should have interesting variables ...
# - MQTT: statistics / variables ?
# - Host / Containers (root ProcessManager)
# - Registrar(s), LifeCycleManager(s), StorageManager(s), HyperSpace
# - Pipeline(s) / PipelineElement(s)
# - Ray node(s)
#
# Ideas
# ~~~~~
# - Dashboard plug-ins: Metrics for all Services/Actors/Agents/Pipelines
#   - LISP REPL: Interactive, composable GUI IDE Graph WorkSpace
#     - Interact with Ollama(Actor), timg (videos, images)
#     - do_discovery(), do_command(), do_request() with Services
#   - mosquitto (MQTT): Metrics (charts), publish, subscribe, DynSec admin
#   - Registrar: Query the Services OrderedDict or as a tree viewer
#   - ProcessManager: LISP REPL using API
#   - HyperSpace: Tree view and LISP REPL using API

import click
from collections import defaultdict, deque
from subprocess import Popen
import pyperclip  # Clipboard support

from asciimatics.event import KeyboardEvent
from asciimatics.exceptions import (
    NextScene, ResizeScreenError, StopApplication
)
from asciimatics.parsers import AnsiTerminalParser
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.widgets import Frame as asciimatics_Frame  # avoid confusion !
from asciimatics.widgets import (
    Label, Layout, MultiColumnListBox,
    PopUpDialog, PopupMenu, Text, TextBox, Widget
)
from asciimatics.widgets.utilities import THEMES

import aiko_services as aiko  # designed to be external to the main framework
from aiko_services.main.connection import ConnectionState
from aiko_services.main.utilities import get_mqtt_configuration, load_module

__all__ = ["LogUI", "ServiceFrame"]

_FRAME_UPDATE_RATE = 20 / 4  # 5 Hz (assuming update rate is roughly 20 Hz)
_HISTORY_LIMIT = 32
_LOG_RING_BUFFER_SIZE = 128
_PLUGINS = {}  # updated by set_plugins()

BLACK = Screen.COLOUR_BLACK
WHITE = Screen.COLOUR_WHITE
GREEN = Screen.COLOUR_GREEN
FONT_BOLD = Screen.A_BOLD
FONT_NORMAL = Screen.A_NORMAL

NICE_COLORS = defaultdict(lambda: (WHITE, FONT_NORMAL, BLACK))
NICE_COLORS["focus_button"] = (GREEN, FONT_BOLD, BLACK)
NICE_COLORS["selected_focus_field"] = (GREEN, FONT_BOLD, BLACK)
NICE_COLORS["title"] = (BLACK, FONT_BOLD, WHITE)
THEMES["nice"] = NICE_COLORS

class FrameCommon:
    def __init__(self, screen, height, width, has_border, name):
        super(FrameCommon, self).__init__(
            screen, height, width, has_border=has_border, name=name)
        self.adjust_palette_required = True

        mqtt_configuration = get_mqtt_configuration()
        server_up = mqtt_configuration[0]
        self.mqtt_host = mqtt_configuration[1]
        self.mqtt_port = mqtt_configuration[2]
        if not server_up:
            self.mqtt_host = "MQTT SERVER UNAVAILABLE"

        self.connection_popup = None
        self.connection_state_last = None
    #   aiko.process.connection.add_handler(self._connection_state_handler)

    def _add_service_bar(self):
        layout = Layout([1])
        self.add_layout(layout)
        service_title = Text()
        service_title.custom_colour = "title"
        service_title.disabled = True
        layout.add_widget(service_title)
        return service_title

    def _get_title(self, name=None, context=""):
        if name:
            title = f"{name}: "
        else:
            title = "AikoServices: "
        if context:
            title += context
        else:
            title += f"{self.mqtt_host[0:40]}:{self.mqtt_port}"
        return title

    def _add_title_bar(self):
        layout = Layout([3, 1])
        self.add_layout(layout)
        layout.add_widget(Label(self._get_title()), 0)
        layout.add_widget(Label('Press "?" for help', align=">"), 1)

    def _adjust_palette(self):
        self.palette = NICE_COLORS
        self.adjust_palette_required = False

    BLACK = 0
    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6
    WHITE = 7

    # Reset: \x1b[0m  Color: \x1b[3?n

    def _color_text(self, color, text):
        return f"\x1b[3{color}m{text}"

    def _connection_state_handler(self, connection, connection_state):
        def _on_close(button_index):
            self.connection_popup = False

        if self.scene:
            self.connection_state_last = connection_state
            if connection.is_connected(ConnectionState.REGISTRAR):
                if self.connection_popup:
                      self.scene.remove_effect(self.connection_popup)
                      self.connection_popup = False
                raise NextScene("Dashboard")
            elif not self.connection_popup:
                self.connection_popup = PopUpDialog(self._screen,
                    "MQTT server and/or Aiko Registar not found",
                    ["OK"], on_close=_on_close, theme="nice")
                self.scene.add_effect(self.connection_popup)

    @property
    def frame_update_count(self):
        return _FRAME_UPDATE_RATE

    def _short_name(self, path):
        after_slash = path.rfind("/") + 1
        return path[after_slash:]

    def _update(self, frame_no):
        super(FrameCommon, self)._update(frame_no)
        if frame_no > 5 and frame_no % _FRAME_UPDATE_RATE:  # roughly 1 Hz
            connection = aiko.process.connection
            connection_state = connection.connection_state
            if self.connection_state_last != connection_state:
                self._connection_state_handler(connection, connection_state)

    def _update_ecproducer_variable(self, topic_path, name, value):
        topic_path_control = topic_path + "/control"
        payload_out = f"(update {name} {value})"
        aiko.process.message.publish(topic_path_control, payload_out)

# TODO: Replace _process_event_common() with the following ...
# https://asciimatics.readthedocs.io/en/stable/widgets.html#global-key-handling
    def _process_event_common(self, event):
        if isinstance(event, KeyboardEvent):
            if event.key_code == ord("?"):
            # TODO: [h] History / Log multiple Services (toggle)
            # TODO: [p] Publish to Service
            # TODO: [K] Kill Service (dialog for terminate / kill, remote)
            #       [Delete] As above
            # TODO: [L] Show Log page: Support multiple Service logs
            # TODO: [N] New Service / Actor / Pipeline (parameters)
            # TODO: [S] Show Service page: Generic page, log level, subscribe
            #                              Parameters
            #       [Enter] As above
            # TODO: [T] Terminate Actor
                text = " Help   Aiko Services Dashboard\n"        \
                       " ----   -----------------------\n"        \
                       " Enter  Update variable value\n"          \
                       " Tab    Move to next section\n"           \
                       " c      Copy topic path to clipboard \n"  \
                       " f      Filter Service (toggle)\n"        \
                       " l      Log level change\n"               \
                       " s      Select Service (toggle)\n"        \
                       " C      Clear selection\n"                \
                       " D      Show Dashboard page\n"            \
                       " K      Kill Service (local)\n"           \
                       " L      Show Log page\n"                  \
                       " S      Show Service page\n"              \
                       " x      Exit"
                self.scene.add_effect(
                    PopUpDialog(self._screen, text, ["OK"], theme="nice"))
            if event.key_code in [ord("x"), ord("X"), Screen.ctrl("c")]:
                self.services_cache = None
                aiko.services_cache_delete()
                raise StopApplication("User quit request")

    def _update_field(self, list, name, value, width):  # wrap long lines
        value = str(value)
        padding = 0
        while len(value):
            if isinstance(name, str):
                field = (name, " "*padding + value[0:width-padding])
                name = ""
            else:
                field = (" "*padding + value[0:width-padding],)
            list.append(field)
            value = value[width-padding:]
            if padding == 0:
                padding = 2

class DashboardFrame(FrameCommon, asciimatics_Frame):
    dashboard_singleton = None  # Instance reference for ServiceFrames plugins

    FILTER_OUT = set(["category", "pipeline_element"])

    @classmethod
    def get_singleton(cls):
        return DashboardFrame.dashboard_singleton

    def __init__(self, screen, _):
        if not DashboardFrame.dashboard_singleton:
            DashboardFrame.dashboard_singleton = self

        super(DashboardFrame, self).__init__(
            screen, screen.height, screen.width, has_border=False,
            name="dashboard_frame"
        )

        self.ec_consumer = None
        self._ec_consumer_reset()
        self.filter_out = DashboardFrame.FILTER_OUT.copy()
        self.services_row = -1
        self.selected_service = None
        self.selected_services = {}
        self.subscribed_service = None

        self.services_cache = aiko.services_cache_create_singleton(
            aiko.process, True, history_limit=_HISTORY_LIMIT)
    #   filter = aiko.ServiceFilter("*", "*", "*", "*", "*", "*")

        self._services_widget = MultiColumnListBox(
            screen.height // 3,
            ["<28", "<14", "<8", "<20", ">0"],  # Transport: <10 columns
            options=[],
            titles=["Service Topic", "Name", "Owner", "Protocol", "Transport"],
            on_change=self._on_change_services,
            parser=AnsiTerminalParser()
        )
        self._service_widget = MultiColumnListBox(
            screen.height // 2,
            ["<24", "<0"],
            options=[],
            titles=["Variable name", "Value"],
            on_select=self._on_select_variable
        )
        self._history_widget = MultiColumnListBox(  # Transport: <10 columns
            Widget.FILL_FRAME,
            ["<28", "<14", "<8", "<20", ">0"],
            options=[],
            titles=["Service history: Topic",
                    "Name", "Owner", "Protocol", "Transport"]
        )
        self._add_title_bar()
        layout_0 = Layout([1], fill_frame=True)
        self.add_layout(layout_0)
        layout_0.add_widget(self._services_widget)
        layout_0.add_widget(self._service_widget)
        layout_0.add_widget(self._history_widget)
        self.fix()  # Prepare asciimatics_Frame for use
        self._value_width = self._service_widget.width - 16

    def _ec_consumer_set(self, index):
        self._ec_consumer_reset()
        self.services_row = -1
        self.selected_service = None

        services = self.services_cache.get_services()
        services_topic_paths = services.get_topic_paths()
        if len(services_topic_paths) > index:
            self.services_row = index
            service_topic_path = services_topic_paths[index]
            self.selected_service = services.get_service(service_topic_path)
            self.service_tags = self.selected_service[5]
            if aiko.ServiceTags.match_tags(self.service_tags, ["ec=true"]):
                topic_control = f"{service_topic_path}/control"
                self.ec_consumer = aiko.ECConsumer(
                    aiko.process, 0, self.service_cache, topic_control)

    def _ec_consumer_reset(self):
        if self.ec_consumer:
            self.ec_consumer.terminate()
        self.ec_consumer = None
        self.service_cache = {}
        self.service_tags = None

    def _filter(self, parent, service, topic_path, protocol):
        protocol = protocol.split(":")[0]
        if topic_path.service_id == "1":
            parent = (topic_path, protocol)

        show = protocol not in self.filter_out
        if topic_path.service_id != "1":
            if "sid>1" in self.filter_out:
                show = False
            if "pipeline_element" in self.filter_out and  \
                parent[1] and parent[1] == "pipeline":
                show = False
        return parent, show

    def _kill_service(self, service):
        topic_path = service[0]
        transport = service[3]
        kill_signal = "-2" if transport == "ray" else "-9"
        if topic_path.count("/") == 3:
            process_id = aiko.ServiceTopicPath.parse(topic_path).process_id
            if process_id.isnumeric():
                command_line = ["kill", kill_signal, process_id]
                Popen(command_line, bufsize=0, shell=False)
                self._ec_consumer_set(self.services_row + 1)

    def _on_change_services(self):
        row = self._services_widget.value
        if row is not None and row != self.services_row:
            self._ec_consumer_set(row)

    def _on_select_variable(self):
        text_box = TextBox(1, None, None, False, False)
        variable_name = None

        def _on_close(button_index):
            if button_index == 1:
                self._update_ecproducer_variable(
                    self.selected_service[0], variable_name, text_box.value[0])

        row = self._service_widget.value
        variable = self._service_widget.options[row]
        variable_name = variable[0][0]
        if len(variable_name) > 0:
            if variable_name != "Tag:" and not variable_name.endswith(" ..."):
                variable_value = variable[0][1]
                text_box.value[0] = variable_value
                title = f"Update {variable_name}" + " "*32
            # TODO: Replace with "maps.py:EnterLocation(asciimatics_Frame)"
                popup_dialog = PopUpDialog(
                    self._screen, title, ["Cancel", "OK"],
                    on_close=_on_close, theme="nice")
                layout = Layout([1])
                popup_dialog.add_layout(layout)
                layout.add_widget(text_box)
                popup_dialog.fix()
                self.scene.add_effect(popup_dialog)

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if event.key_code == ord("c") and self.selected_service:
                try:
                    pyperclip.copy(self.selected_service[0])
                except pyperclip.PyperclipException:
                    text = 'Clipboard copy/paste failed\n'          \
                           'Python package "pyperclip" requires\n'  \
                           'Linux package  "xclip"'
                    self.scene.add_effect(
                        PopUpDialog(self._screen, text, ["OK"], theme="nice"))
            if event.key_code == ord("f"):
                self.scene.add_effect(FilterPopupMenu(self._screen,
                    self._services_widget))
            if event.key_code == ord("l") and self.selected_service:
                self.scene.add_effect(LogLevelPopupMenu(self._screen,
                    self._services_widget, self.selected_service[0]))
            if event.key_code == ord("s") and self.selected_service:
                self._service_selection_toggle(self.selected_service)
            if event.key_code == ord("C"):
                self._service_selection_clear()
            if event.key_code == ord("K") and self.selected_service:
                self._kill_service(self.selected_service)
            if event.key_code == ord("L") and self.selected_service:
                raise NextScene("Log")
            if event.key_code == ord("S") and self.selected_service:
                self._raise_next_scene(self.selected_service)
        self._process_event_common(event)
        return super(DashboardFrame, self).process_event(event)

    def _raise_next_scene(self, service):
        service_name = service[1]
        service_protocol = self._short_name(service[2]).split(":")[0]
        names = [service_name, service_protocol]
        scene_name = [name for name in names if name in _PLUGINS]

        if scene_name:
            raise NextScene(scene_name[0])
        else:
            text = f" Service: {service_name} \n Does not have a custom page "
            self.scene.add_effect(
                PopUpDialog(self._screen, text, ["OK"], theme="nice"))

    def _service_selection_clear(self):
        self.selected_services = {}

    def _service_selection_color(self, color, topic_path, topic_path_show):
        topic_path_colored = topic_path_show
        if topic_path in self.selected_services:
            topic_path_colored = self._color_text(color, topic_path_show)
        return topic_path_colored

    def _service_selection_toggle(self, service):
        topic_path = service[0]
        if topic_path in self.selected_services:
            del self.selected_services[topic_path]
        else:
            self.selected_services[topic_path] = service

    def _update(self, frame_no):
        if self.adjust_palette_required:
            self._adjust_palette()

        parent = (None, None)
        services_count = self.services_cache.get_services().count   # correct
        services = self.services_cache.get_services().copy()  # count is zero
        services_formatted = []
        for service in services:
            topic_path = aiko.ServiceTopicPath.parse(service[0])
            topic_color = self._service_selection_color(
                self.YELLOW, str(topic_path), topic_path.terse)
            protocol = self._short_name(service[2])
            parent, show = self._filter(parent, service, topic_path, protocol)
            if show:
                services_formatted.append(
                    (topic_color, service[1], service[4], protocol, service[3]))
        self._services_widget.options = [
            (service_info, row_index)
            for row_index, service_info in enumerate(services_formatted)
        ]

        variables = []
        if self.ec_consumer:
            service_variables = list(self.service_cache.items())
            for variable_name, variable_value in service_variables:
                if isinstance(variable_value, dict):
                    variables.append((f"{variable_name} ...", ""))
                    for name, value in variable_value.items():
                        variables.append((f"  {name}", f"{value}"))
                else:
                #   variables.append((variable_name, variable_value))
                    self._update_field(
                        variables, variable_name,
                        variable_value, self._value_width)
            variables.append(("", ""))

        if self.service_tags:
            for service_tag in self.service_tags:
            #   variables.append(("Tag:", service_tag))
                self._update_field(
                    variables, "Tag:", service_tag, self._value_width)

        self._service_widget.options = [
            (variable, row_index)
            for row_index, variable in enumerate(variables)
        ]

        parent = (None, None)
        service_history = list(self.services_cache.get_history())
        services_formatted = []
        for service in service_history:
            topic_path = aiko.ServiceTopicPath.parse(service[0])
            topic_terse = topic_path.terse
            protocol = self._short_name(service[2])
            parent, show = self._filter(parent, service, topic_path, protocol)
            if show:
                services_formatted.append(
                    (topic_terse, service[1], service[4], protocol, service[3]))
        self._history_widget.options = [
            (service_info, row_index)
            for row_index, service_info in enumerate(services_formatted)
        ]

        super(DashboardFrame, self)._update(frame_no)

# ServiceFrame subclass __init__() must include "self.fix()"

class ServiceFrame(FrameCommon, asciimatics_Frame):
    def __init__(self, screen, dashboard, name="service_frame"):
        super(ServiceFrame, self).__init__(
            screen, screen.height, screen.width, has_border=False, name=name
        )

        self.dashboard = dashboard
        self.service = None
        self._add_title_bar()
        self._service_title = self._add_service_bar()
    #   self._value_width = self._service_widget.width

    def _update(self, frame_no):
        if self.adjust_palette_required:
            self._adjust_palette()

        if self.dashboard.selected_service != self.dashboard.subscribed_service:
            self.dashboard.subscribed_service = self.dashboard.selected_service
            self.service = self.dashboard.selected_service
            topic_path = aiko.ServiceTopicPath.parse(self.service[0]).terse
            name = self._short_name(self.service[1])
            self._service_title.value = f"Service: {topic_path}: {name}"
        # Plugins that add handlers to the ec_consumer must also remove them !
            self._service_frame_start(self.service, self.dashboard.ec_consumer)

        super(ServiceFrame, self)._update(frame_no)

    def _service_frame_start(self, service, service_ec_consumer):
        pass

    def _service_frame_stop(self, service):
        pass

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if event.key_code == ord("D"):
                self._service_frame_stop(self.service)
                self.dashboard.subscribed_service = None
                raise NextScene("Dashboard")
        self._process_event_common(event)
        return super(ServiceFrame, self).process_event(event)

# Re-usable logging user interface  # TODO: Consider turning into a Widget ?
# https://asciimatics.readthedocs.io/en/stable/widgets.html#custom-widgets

class LogUI:
    def __init__(self, parent, height=Widget.FILL_FRAME):
        self.parent = parent
        self.log_buffer = None
        self.log_show_latest = True
        self.recorder = None
        self.topic_paths = {}

        self._log_widget = MultiColumnListBox(
            height, ["<0"], options=[],
            titles=["Date       Time        Log level Message"]
        )
        layout_0 = Layout([1])
        parent.add_layout(layout_0)
        recorder_label = Label(f"Log Recorder: TOPIC_PATH")
        recorder_label.custom_colour = "title"
        layout_0.add_widget(recorder_label)
        layout_1 = Layout([1], fill_frame=True)
        parent.add_layout(layout_1)
        layout_1.add_widget(self._log_widget)

    def _add_topic_path(self, topic_path):
        if topic_path not in self.topic_paths:
            self.topic_paths[topic_path] = topic_path
            topic_path, _, _ = topic_path.rpartition("/")
            topic_path += "/0/log"  # TODO: Use correct Service Id
            aiko.process.add_message_handler(
                self._topic_log_handler, topic_path)

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            widget = self._log_widget
            if len(widget._options) > 0 and event.key_code == Screen.KEY_HOME:
                self.log_show_latest = False
                widget._line = 0
                widget.value = widget.options[widget._line][1]
            if len(widget._options) > 0 and event.key_code == Screen.KEY_END:
                self.log_show_latest = True
                widget._line = len(widget._options) - 1
                widget.value = widget.options[widget._line][1]
            if event.key_code in [Screen.KEY_UP, Screen.KEY_PAGE_UP]:
                self.log_show_latest = False
            if event.key_code in [Screen.KEY_DOWN, Screen.KEY_PAGE_DOWN]:
                if widget._line == len(widget._options) - 1:
                    self.log_show_latest = True

    def _remove_topic_path(self, topic_path):
        if topic_path in self.topic_paths:
            del self.topic_paths[topic_path]
            topic_path, _, _ = topic_path.rpartition("/")
            topic_path += "/0/log"  # TODO: Use correct Service Id
            aiko.process.remove_message_handler(
                self._topic_log_handler, topic_path)

    def _service_frame_start(self, service, service_ec_consumer):
        self.log_buffer = deque(maxlen=_LOG_RING_BUFFER_SIZE)
        self._add_topic_path(service[0])

    def _service_frame_stop(self, service):
        topic_paths = self.topic_paths.copy()
        for topic_path in topic_paths:
            self._remove_topic_path(topic_path)
        self.log_buffer = None
        self.topic_paths = {}

    def _topic_log_handler(self, _aiko, topic, payload_in):
        self.log_buffer.append(payload_in)

    def _update(self, frame_no):
        widget = self._log_widget
        if self.log_show_latest and len(widget._options) > 0:
            widget._line = len(widget._options) - 1
            widget.value = widget.options[widget._line][1]

        log_records = []
        if self.log_buffer:
            for log_record in self.log_buffer:
            #   log_records.append((log_record,))  # No record line wrapping
                self.parent._update_field(         # With record line wrapping
                    log_records, None, log_record, self._log_widget.width)
        self._log_widget.options = [
            (log_record, row_index)
            for row_index, log_record in enumerate(log_records)
        ]

class LogFrame(ServiceFrame):
    def __init__(self, screen, dashboard):
        super(LogFrame, self).__init__(screen, dashboard, name="log_frame")
        self.log_ui = LogUI(self)
        self.fix()  # Prepare asciimatics_Frame for use

    def process_event(self, event):
        self.log_ui.process_event(event)
        return super(LogFrame, self).process_event(event)

    def _service_frame_start(self, service, service_ec_consumer):
        self.log_ui._service_frame_start(service, service_ec_consumer)

    def _service_frame_stop(self, service):
        self.log_ui._service_frame_stop(service)

    def _update(self, frame_no):
        super(LogFrame, self)._update(frame_no)
        self.log_ui._update(frame_no)

class FilterPopupMenu(PopupMenu):
    def __init__(self, screen, parent_widget):
        self._dashboard = DashboardFrame.get_singleton()
        self._screen = screen
        self._parent_widget = parent_widget

        menu_items = [
            ("   Cancel",          self._parent_widget.focus),
            ("1: Service id > 1",  self._button_handler),
            ("c: Category",        self._button_handler),
            ("e: PipelineElement", self._button_handler),
            ("R: Reset",           self._button_handler)]
        x = screen.width // 2 - 4
        y = screen.height // 3

        super().__init__(self._screen, menu_items, x, y)
        self.palette = NICE_COLORS

    def _button_handler(self):
        log_level = self.focussed_widget.text.upper()
    #   self._set_log_level(log_level)

    def _set_log_level(self, log_level):
    #   self._dashboard._update_ecproducer_variable(
    #       self._service_selected, "log_level", log_level)
        self._parent_widget.focus()

    filter_key_map = {
        ord("1"): "sid>1",
        ord("c"): "category",
        ord("e"): "pipeline_element"
    }

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if event.key_code == ord("R"):
                self._dashboard.filter_out = DashboardFrame.FILTER_OUT.copy()
                self._destroy()
            if event.key_code in FilterPopupMenu.filter_key_map:
                filter = FilterPopupMenu.filter_key_map[event.key_code]
                if filter in self._dashboard.filter_out:
                    self._dashboard.filter_out.discard(filter)
                else:
                    self._dashboard.filter_out.add(filter)
                self._destroy()
        return super(FilterPopupMenu, self).process_event(event)

class LogLevelPopupMenu(PopupMenu):
    def __init__(self, screen, parent_widget, service_selected):
        self._dashboard = DashboardFrame.get_singleton()
        self._screen = screen
        self._parent_widget = parent_widget
        self._service_selected = service_selected

        menu_items = [
            ("   Cancel",      self._parent_widget.focus),
            ("d: Debug",       self._button_handler),
            ("e: Error",       self._button_handler),
            ("i: Info",        self._button_handler),
            ("w: Warning",     self._button_handler),
            ("D: Debug   all", self._button_handler),
            ("E: Error   all", self._button_handler),
            ("I: Info    all", self._button_handler),
            ("W: Warning all", self._button_handler)]
        x = screen.width // 2 - 4
        y = screen.height // 3

        super().__init__(self._screen, menu_items, x, y)
        self.palette = NICE_COLORS

    def _button_handler(self):
        log_level = self.focussed_widget.text.upper()
        self._set_log_level(log_level)

    def _set_log_level(self, log_level):
        self._dashboard._update_ecproducer_variable(
            self._service_selected, "log_level", log_level)
        self._parent_widget.focus()

    log_level_key_map = {
        ord("d"): "DEBUG",
        ord("e"): "ERROR",
        ord("i"): "INFO",
        ord("w"): "WARNING",
        ord("D"): "DEBUG_ALL",
        ord("E"): "ERROR_ALL",
        ord("I"): "INFO_ALL",
        ord("W"): "WARNING_ALL"
    }

    def process_event(self, event):
        if isinstance(event, KeyboardEvent):
            if self._dashboard.selected_service:
                if event.key_code in LogLevelPopupMenu.log_level_key_map:
                    level = LogLevelPopupMenu.log_level_key_map[event.key_code]
                    self._set_log_level(level)
                    self._destroy()
        return super(LogLevelPopupMenu, self).process_event(event)

def run_dashboard(screen, start_scene):
    scenes = []
    dashboard = None
    for scene_name, scene_class in _PLUGINS.items():
        scene = Scene([scene_class(screen, dashboard)], -1, name=scene_name)
        if scene_name == "Dashboard":
            dashboard = DashboardFrame.get_singleton()
        scenes.append(scene)
    screen.play(scenes, stop_on_resize=True, start_scene=start_scene)

def update_plugins(plugins):
    if not _PLUGINS:
        _PLUGINS["Dashboard"] = DashboardFrame
        _PLUGINS["Log"] = LogFrame
    _PLUGINS.update(plugins)

@click.command()
@click.option("--history_limit", "-hl", type=click.INT, default=32,
    help="History length requested from Registrar")
@click.option("--plugin", "-p", multiple=True, required=False,
    default=["aiko_services.main.dashboard_plugins"])

def main(plugin, history_limit):
    global _HISTORY_LIMIT
    _HISTORY_LIMIT = history_limit

    for plugin_name in plugin:
        try:
            module = load_module(plugin_name) if plugin_name else None
        except ModuleNotFoundError:
            module = None
        update_plugins(module.plugins if module else {})

    scene = None
    while True:
        try:
            Screen.wrapper(
                run_dashboard, catch_interrupt=True, arguments=[scene]
            )
            break
        except ResizeScreenError as exception:
            scene = exception.scene

if __name__ == "__main__":
    main()



import os
import tomllib
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union

import tomli_w
from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    ValidationError,
    field_validator,
    model_validator,
)

from .const import (
    HL_DIR,
)


class HighlighterRuntimeConfigError(Exception):
    """Raised when there’s a problem loading or validating the Highlighter Runtime config file."""

    pass


class HighlighterRuntimeConfig(BaseModel):
    _instance: "HighlighterRuntimeConfig | None" = None

    class AgentConfig(BaseModel):

        _local_cache_directory: Path = PrivateAttr()

        queue_response_max_size: int = Field(
            default=100,
            ge=0,
            description="Maximum size of an agent's response queue. The larger the queue, the more memory consumed. Defaults to 100.",
        )
        timeout_secs: float = Field(
            default=120.0,
            ge=0.0,
            description="Timeout in seconds when the agent is processing data. Defaults to 120s. For example, when the timeout is 30s, if there is no output from the agent after 30s, a timeout error is raised.",
        )
        task_lease_duration_secs: float = Field(
            default=60.0,
            ge=0.0,
            description="Duration in seconds to lease a task for processing by an agent. Default is 60s.",
        )
        task_polling_period_secs: float = Field(
            default=5.0,
            ge=0.0,
            description="Period in seconds to poll a task for processing by an agent. Default is to poll every 5s.",
        )
        memory_gc_every_n_frames: int = Field(
            default=0,
            ge=0,
            description="Run gc.collect every N frames. Set to 0 to disable.",
        )
        memory_gc_every_n_seconds: float = Field(
            default=0.0,
            ge=0.0,
            description="Run gc.collect every N seconds. Set to 0 to disable.",
        )
        memory_malloc_trim_every_n_frames: int = Field(
            default=0,
            ge=0,
            description="Run malloc_trim every N frames (Linux/glibc only). Set to 0 to disable.",
        )
        memory_malloc_trim_every_n_seconds: float = Field(
            default=0.0,
            ge=0.0,
            description="Run malloc_trim every N seconds (Linux/glibc only). Set to 0 to disable.",
        )

        def agents_dir(self) -> Path:
            # FIXME: Should this be nested under accounts_dir too?
            pth = Path(self._local_cache_directory) / "agents"
            pth.mkdir(parents=True, exist_ok=True)
            return pth

        def data_dir(self) -> Path:
            # FIXME: See comment in self.agents_dir
            pth = Path(self.agents_dir()) / "data"
            pth.mkdir(parents=True, exist_ok=True)
            return pth

        def db_dir(self) -> Path:
            # FIXME: See comment in self.agents_dir
            pth = Path(self.agents_dir()) / "db"
            pth.mkdir(parents=True, exist_ok=True)
            return pth

        def db_file(self) -> Path:
            # FIXME: See comment in self.agents_dir
            pth = Path(self.db_dir() / "database.sqlite")
            return pth

    agent: AgentConfig = Field(default_factory=AgentConfig)

    class NetworkConfig(BaseModel):
        reset_timeout: int = Field(
            default=60,
            description="Time in seconds that the circuit breaker remains open before transitioning to half-open and retrying.",
        )
        max_retries: int = Field(
            default=3,
            description="Maximum number of retry attempts for each network operation before giving up.",
        )

    network: NetworkConfig = Field(default_factory=NetworkConfig)

    class OnnxRuntimeConfig(BaseModel):
        shared_allocator: bool = Field(
            default=False,
            description="Share a single CPU arena allocator across sessions. Defaults to False.",
        )
        enable_cpu_mem_arena: bool = Field(
            default=True,
            description="Enable the CPU memory arena allocator. Defaults to True.",
        )
        arena_extend_strategy: Optional[int] = Field(
            default=None,
            description=(
                "CPU arena extend strategy. Accepts 0/1 or aliases " "('power2', 'same_as_requested')."
            ),
        )
        arena_max_mem_bytes: Optional[int] = Field(
            default=None,
            ge=0,
            description="Maximum CPU arena memory in bytes (optional).",
        )
        arena_initial_chunk_bytes: Optional[int] = Field(
            default=None,
            ge=0,
            description="Initial CPU arena chunk size in bytes (optional).",
        )
        use_global_threadpool: bool = Field(
            default=False,
            description="Use global/shared threadpools for ORT sessions. Defaults to False.",
        )
        session_cache_maxsize: int = Field(
            default=4,
            ge=1,
            description="Maximum number of cached ORT sessions. Defaults to 4.",
        )
        arena_shrinkage: Optional[str] = Field(
            default=None,
            description=(
                "Enable arena shrinkage per Run() by setting a device list "
                "(e.g., 'cpu:0', 'gpu:0', or 'cpu:0;gpu:0')."
            ),
        )
        cuda_arena_extend_strategy: Optional[Literal["kNextPowerOfTwo", "kSameAsRequested"]] = Field(
            default=None,
            description="CUDA arena extend strategy for the CUDAExecutionProvider.",
        )
        cuda_gpu_mem_limit_bytes: Optional[int] = Field(
            default=None,
            ge=0,
            description="CUDA GPU memory limit in bytes for the CUDAExecutionProvider.",
        )

        @field_validator("cuda_arena_extend_strategy", mode="before")
        @classmethod
        def normalize_cuda_arena_extend_strategy(cls, value: Any):
            if value is None:
                return None
            if isinstance(value, str):
                normalized = value.strip()
                lowered = normalized.lower()
                if lowered in {"knextpoweroftwo", "next_power_of_two", "power2"}:
                    return "kNextPowerOfTwo"
                elif lowered in {"ksameasrequested", "same_as_requested"}:
                    return "kSameAsRequested"
            raise ValueError(
                "cuda_arena_extend_strategy must be one of: " "kNextPowerOfTwo, kSameAsRequested"
            )

        @field_validator("arena_extend_strategy", mode="before")
        @classmethod
        def normalize_arena_extend_strategy(cls, value: Any):
            if value is None:
                return None
            if isinstance(value, int):
                if value in (0, 1):
                    return value
            if isinstance(value, str):
                normalized = value.strip()
                lowered = normalized.lower()
                if lowered in {"power2", "knextpoweroftwo", "next_power_of_two", "0"}:
                    return 0
                if lowered in {"same_as_requested", "ksameasrequested", "1"}:
                    return 1
            raise ValueError("arena_extend_strategy must be one of: 0, 1, power2, same_as_requested")

    onnxruntime: OnnxRuntimeConfig = Field(default_factory=OnnxRuntimeConfig)

    @model_validator(mode="after")
    def sync_root_dir(self):
        # Pass the local_cache_directory defined in the "outer" part
        # of the BaseModel to the "inner" AgentConfig BaseModel
        self.agent._local_cache_directory = Path(self.local_cache_directory)
        return self

    ### Client level config
    config_path: Optional[str] = Field(
        default=None,
        description="Location that the config was loaded from",
    )

    # Do we always want Path or str or should we accept both
    local_cache_directory: str = Field(
        default=str(HL_DIR),
        description="Local cache directory used by agents and SDK. Defaults to $HOME/.highlighter.",
    )

    download_timeout_secs: float = Field(
        default=300.0, ge=0.0, description="Timeout in seconds when downloading data. Defaults to 300s."
    )

    log_path: str = Field(
        default=str(HL_DIR / "log" / "development.log"),
        description="Path to log file used by agents and SDK. Defaults to $HOME/.highlighter/log/development.log",
    )

    log_level: str = Field(
        default="WARNING",
        description="Log level used by agents and SDK. Set to one of python log levels: DEBUG, INFO, WARNING, ERROR, CRITICAL. Defaults to WARNING.",
    )

    log_rotation_max_kilobytes: int = Field(
        default=100 * 1024,  # 100 MB
        ge=1,  # Minimum 1KB
        description="Maximum size of a log file in kilobytes before rotation. Defaults to 100MB (102400 KB).",
    )

    log_rotation_backup_count: int = Field(
        default=4,
        ge=0,
        description="Number of backup log files to keep when rotating. Defaults to 4 (5 total files including current).",
    )

    pagination_page_size: int = Field(
        default=200,
        ge=1,
        description="Number or responses per page when paginating GraqpQL connection queries",
    )

    def account_dir(self, account_name: str) -> Path:
        pth = Path(self.local_cache_directory) / "accounts" / account_name
        pth.mkdir(parents=True, exist_ok=True)
        return pth

    def data_sources_dir(self, account_name: str) -> Path:
        pth = self.account_dir(account_name) / "data_sources"
        pth.mkdir(parents=True, exist_ok=True)
        return pth

    def data_files_dir(self, account_name: str) -> Path:
        pth = self.account_dir(account_name) / "data_files"
        pth.mkdir(parents=True, exist_ok=True)
        return pth

    def datasets_dir(self, account_name: str) -> Path:
        pth = self.account_dir(account_name) / "datasets"
        pth.mkdir(parents=True, exist_ok=True)
        return pth

    def data_models_dir(self, account_name: str) -> Path:
        # pth = self.account_dir(account_name) / "data_models"
        # FIXME: data_modals -> data_models
        pth = self.account_dir(account_name) / "data_modals"
        pth.mkdir(parents=True, exist_ok=True)
        return pth

    @property
    def highlighter_config_path(self) -> str:
        """Location to save the Highlighter configuration file."""
        return str(Path(self.local_cache_directory) / "config")

    @classmethod
    def reset_cache(cls) -> None:
        """Clear the cached HighlighterRuntimeConfig instance."""
        cls._instance = None

    @classmethod
    def load(
        cls,
        config_path: Optional[str] = None,
        *,
        force_reload: bool = False,
    ) -> "HighlighterRuntimeConfig":
        """
        Load configuration from the specified path or the default path.

        If the configuration file doesn't exist:
        - If using the default path, create it with default values
        - If using a custom path, notify the user and exit

        Args:
            config_path: Path to the configuration file or None to use the default

        Returns:
            HighlighterRuntimeConfig: The loaded configuration
        """
        if isinstance(cls._instance, cls) and not force_reload:
            return cls._instance

        default_path = (HL_DIR / "config").expanduser()

        if config_path is None:
            path = default_path
        else:
            path = Path(config_path).expanduser()

        raw_config = parse_toml_file(path)

        if raw_config is None:
            if path == default_path:
                default_config = cls()
                write_default_config(path, default_config)
                default_config.override_with_env_vars()
                default_config.config_path = str(path)
                cls._instance = default_config
                return default_config
            else:
                raise HighlighterRuntimeConfigError(
                    f"Config file not found at {path}. Use --config to specify a valid configuration file or run without --config to use defaults."
                )
        try:
            raw_config["config_path"] = str(path)
            cfg = cls(**raw_config)
            cfg.override_with_env_vars()
            cls._instance = cfg
            return cfg

        except ValidationError as e:
            raise HighlighterRuntimeConfigError(f"Invalid configuration '{path}': {e}")

    def override_with_env_vars(self) -> None:
        """
        Override configuration values if corresponding environment variables are set.

        Supported environment variables:
        - HL_AGENT_QUEUE_RESPONSE_MAX_SIZE (int)
        - HL_AGENT_MEM_GC_EVERY_N_FRAMES (int)
        - HL_AGENT_MEM_GC_EVERY_N_SECONDS (float)
        - HL_AGENT_MEM_MALLOC_TRIM_EVERY_N_FRAMES (int)
        - HL_AGENT_MEM_MALLOC_TRIM_EVERY_N_SECONDS (float)
        - HL_CACHE_DIR (str)
        - HL_DOWNLOAD_TIMEOUT (int)
        - HL_LOG_LEVEL (str)
        """
        # Agent queue size
        val = os.getenv("HL_AGENT_QUEUE_RESPONSE_MAX_SIZE")
        if val is not None:
            try:
                self.agent.queue_response_max_size = int(val)
            except ValueError:
                raise ValueError(f"HL_AGENT_QUEUE_RESPONSE_MAX_SIZE must be an integer, got '{val}'")

        env_configs = [
            ("HL_AGENT_MEM_GC_EVERY_N_FRAMES", "memory_gc_every_n_frames", int, "integer"),
            ("HL_AGENT_MEM_GC_EVERY_N_SECONDS", "memory_gc_every_n_seconds", float, "float"),
            ("HL_AGENT_MEM_MALLOC_TRIM_EVERY_N_FRAMES", "memory_malloc_trim_every_n_frames", int, "integer"),
            (
                "HL_AGENT_MEM_MALLOC_TRIM_EVERY_N_SECONDS",
                "memory_malloc_trim_every_n_seconds",
                float,
                "float",
            ),
        ]

        for env_var, field_name, cast_func, type_name in env_configs:
            val = os.getenv(env_var)
            if val is not None:
                try:
                    setattr(self.agent, field_name, cast_func(val))
                except ValueError:
                    raise ValueError(f"{env_var} must be a {type_name}, got '{val}'")

        # Local cache directory
        val = os.getenv("HL_CACHE_DIR")
        if val:
            self.local_cache_directory = val

        # Download timeout seconds
        val = os.getenv("HL_DOWNLOAD_TIMEOUT")
        if val is not None:
            try:
                self.download_timeout_secs = float(val)
            except ValueError:
                raise ValueError(f"HL_DOWNLOAD_TIMEOUT must be a float, got '{val}'")

        # Log level
        val = os.getenv("HL_LOG_LEVEL")
        if val is not None:
            self.log_level = val.upper()

        # Keep derived agent paths in sync with cache overrides
        self.agent._local_cache_directory = Path(self.local_cache_directory)

    def save(self, config_path: Optional[str] = None) -> None:
        """
        Save the current configuration to the specified path, or to the default path ~/.highlighter/config if none provided.
        """
        default_path = Path(self.highlighter_config_path).expanduser()
        path = Path(config_path).expanduser() if config_path else default_path
        write_default_config(path, self)


def ensure_directory_exists(path: Union[str, Path]) -> None:
    """Ensure the directory for the given file path exists."""
    directory = Path(path).parent
    try:
        directory.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        raise PermissionError(f"No permission to create directory: {directory}")
    except Exception as e:
        raise Exception(f"Could not create directory {directory}: {e}")


def write_default_config(config_path: Union[str, Path], config: HighlighterRuntimeConfig) -> None:
    """Write a default configuration to the specified path."""
    import logging

    config_path = Path(config_path)
    ensure_directory_exists(config_path)
    logger = logging.getLogger(__name__)

    try:
        with open(config_path, "wb") as f:
            config_dict = (
                config.model_dump()
            )  # Using model_dump() instead of dict() for Pydantic v2 compatibility
            f.write(tomli_w.dumps(config_dict).encode("utf-8"))
        logger.info(f"Created default configuration at: {config_path}")
    except PermissionError:
        logger.warning(f"No permission to write default configuration: {config_path}")
        # Still return the default config even if we can't write it
    except Exception as e:
        logger.warning(f"Unable to write default configuration: {config_path}: {e}")
        # Still return the default config even if we can't write it


def parse_toml_file(file_path: Union[str, Path]) -> Optional[Dict[str, Any]]:
    """Parse a TOML file and return its contents as a dictionary."""
    try:
        with open(file_path, "rb") as f:
            return tomllib.load(f)
    except tomllib.TOMLDecodeError as e:
        raise tomllib.TOMLDecodeError(f"Error parsing TOML in configuration file {file_path}:")
    except FileNotFoundError:
        return None
    except PermissionError:
        raise PermissionError(f"No permission to read configuration: {file_path}")
    except Exception as e:
        raise Exception(f"Error reading configuration {file_path}: {e}")



"""
Case messaging utilities for machine agents.

Provides helpers for agents to send messages to cases during task execution.
"""

import logging
import time
from datetime import datetime, timedelta
from typing import Dict, Optional

from highlighter.client import HLClient
from highlighter.client.base_models.base_models import (
    CreateCaseMessagePayload,
    MessageType,
    TaskType,
)

logger = logging.getLogger(__name__)


class CaseMessageRateLimiter:
    """Rate limiter for case messages to prevent flooding."""

    def __init__(
        self,
        max_messages_per_case: int = 10,
        time_window_seconds: float = 300.0,  # 5 minutes
    ):
        """
        Initialize rate limiter.

        Args:
            max_messages_per_case: Maximum number of messages per case within time window
            time_window_seconds: Time window in seconds for rate limiting
        """
        self.max_messages = max_messages_per_case
        self.time_window = time_window_seconds
        self._message_timestamps: Dict[str, list] = {}

    def check_and_record(self, case_id: str) -> bool:
        """
        Check if a message can be sent to the case and record the timestamp.

        Args:
            case_id: The case ID to check

        Returns:
            True if message can be sent, False if rate limit exceeded
        """
        current_time = time.time()
        cutoff_time = current_time - self.time_window

        # Initialize list for case if not exists
        if case_id not in self._message_timestamps:
            self._message_timestamps[case_id] = []

        # Remove old timestamps outside the time window
        self._message_timestamps[case_id] = [
            ts for ts in self._message_timestamps[case_id] if ts > cutoff_time
        ]

        # Check if under limit
        if len(self._message_timestamps[case_id]) >= self.max_messages:
            return False

        # Record this message timestamp
        self._message_timestamps[case_id].append(current_time)
        return True


class CaseMessenger:
    """Helper for sending messages to cases from machine agents."""

    def __init__(
        self,
        client: Optional[HLClient] = None,
        enabled: bool = True,
        max_messages_per_case: int = 10,
        time_window_seconds: float = 300.0,
    ):
        """
        Initialize case messenger.

        Args:
            client: HLClient instance. If None, will use HLClient.get_client()
            enabled: Whether messaging is enabled
            max_messages_per_case: Maximum messages per case in time window
            time_window_seconds: Time window for rate limiting in seconds (default 5 minutes)
        """
        self.client = client
        self.enabled = enabled
        self.rate_limiter = CaseMessageRateLimiter(max_messages_per_case, time_window_seconds)
        self._no_case_warnings: set = set()  # Track tasks we've warned about

    def _get_client(self) -> HLClient:
        """Lazily acquire the HLClient to avoid requiring credentials at import time."""
        if self.client is None:
            self.client = HLClient.get_client()
        return self.client

    def send_message(
        self,
        task: TaskType,
        content: str,
        force: bool = False,
    ) -> Optional[MessageType]:
        """
        Send a message to the case associated with a task.

        Args:
            task: The task being processed
            content: Message content to send
            force: If True, bypass rate limiting

        Returns:
            The created message if successful, None otherwise
        """
        if not self.enabled:
            logger.debug("Case messaging is disabled")
            return None

        # Check if task has an associated case
        if not task.case or not task.case.id:
            # Log warning once per task
            if task.id not in self._no_case_warnings:
                logger.warning(
                    f"Task {task.id} has no associated case. Cannot send message. "
                    "This warning will only be shown once per task."
                )
                self._no_case_warnings.add(task.id)
            return None

        case_id = task.case.id

        # Check rate limit unless forced
        if not force:
            if not self.rate_limiter.check_and_record(case_id):
                logger.warning(
                    f"Rate limit exceeded for case {case_id}. Message not sent: '{content[:100]}...'"
                )
                return None

        try:
            # Send message via GraphQL mutation
            client = self._get_client()
            result = client.createCaseMessage(
                return_type=CreateCaseMessagePayload,
                caseId=case_id,
                content=content,
            )

            if result.errors:
                logger.error(f"Failed to send message to case {case_id}: {result.errors}")
                return None

            if result.user_message:
                logger.info(f"Sent message to case {case_id}: '{content[:100]}...'")
                return result.user_message
            else:
                logger.warning(f"No message returned after sending to case {case_id}")
                return None

        except Exception as e:
            logger.error(f"Error sending message to case {case_id}: {e}", exc_info=True)
            return None

    def send_status_message(
        self,
        task: TaskType,
        status: str,
        details: Optional[str] = None,
    ) -> Optional[MessageType]:
        """
        Send a status update message to the case.

        Args:
            task: The task being processed
            status: Status to report (e.g., "started", "processing", "completed")
            details: Optional additional details

        Returns:
            The created message if successful, None otherwise
        """
        agent_name = getattr(task, "name", "Agent")
        content = f"[{agent_name}] Status: {status}"
        if details:
            content += f" - {details}"

        return self.send_message(task, content)

    def send_error_message(
        self,
        task: TaskType,
        error: str,
    ) -> Optional[MessageType]:
        """
        Send an error message to the case.

        Args:
            task: The task being processed
            error: Error message to send

        Returns:
            The created message if successful, None otherwise
        """
        agent_name = getattr(task, "name", "Agent")
        content = f"[{agent_name}] Error: {error}"

        # Use force=True for error messages to bypass rate limiting
        return self.send_message(task, content, force=True)


# Global default messenger is created eagerly so instantiation is import-time serialized
_default_messenger = CaseMessenger()


def get_case_messenger() -> CaseMessenger:
    """Return the shared CaseMessenger instance with rate limiting state."""
    return _default_messenger


def send_case_message(
    task: TaskType,
    content: str,
    client: Optional[HLClient] = None,
) -> Optional[MessageType]:
    """
    Convenience function to send a message to a case.

    Note:
        Supplying a custom ``client`` creates a one-off CaseMessenger with its own
        rate limiter state. To reuse rate limiting across calls, rely on the
        shared messenger by omitting the ``client`` argument or manage a dedicated
        ``CaseMessenger`` instance yourself.

    Args:
        task: The task being processed
        content: Message content
        client: Optional HLClient instance

    Returns:
        The created message if successful, None otherwise
    """
    messenger = CaseMessenger(client=client) if client else get_case_messenger()
    return messenger.send_message(task, content)



from .hl_agent_cli_runner import *
from .patch_capability import *
from .rtsp_server import RTSPServer


class FakeBinaryStdin:
    def __init__(self, buf):
        self.buffer = buf  # binary interface

    def read(self, *args, **kwargs):
        return self.buffer.read(*args, **kwargs)

    def isatty(self):
        return False



# To Do
# ~~~~~
# - PackDataSamples / UnpackDataSamples applies to any third-party "raw data"
#   and is not specific to Aiko Services compatibility
#   - Move the "pack" and "unpack" functions into "client/data_sample.py"
#   - This file can then adapt those two functions into Capabilities
#   - Rename this file to something that represents a more general purpose

from datetime import datetime
from typing import Dict, List, Tuple
from uuid import uuid4

from highlighter.agent.capabilities.base_capability import Capability, StreamEvent
from highlighter.core.data_models.data_sample import DataSample

__all__ = ["PackDataSamples", "UnpackDataSamples"]


class PackDataSamples(Capability):

    class StreamParameters(Capability.InitParameters):
        input_key: str
        content_type: str
        output_key: str = "data_samples"
        output_empty_entities: bool = False

    def process_frame(self, stream, *args, **kwargs) -> Tuple[StreamEvent, Dict]:
        parameters = self.stream_parameters(stream.stream_id)
        data_samples = []
        inputs = kwargs[parameters.input_key]
        for input in inputs:
            file_id = uuid4()
            data_sample = DataSample(
                content=input,
                content_type=parameters.content_type,
                data_file_id=file_id,
                recorded_at=datetime.now(),
            )
            data_samples.append(data_sample)

        if "timestamps" in stream.variables:
            timestamps = [datetime.fromtimestamp(ts) for ts in stream.variables["timestamps"]]
        else:
            rate, found = self.pipeline.pipeline_graph.nodes()[0].element.get_parameter("rate")
            if not found:
                raise ValueError("Cannot determine frame 'rate' from head node")

            frame_counter = getattr(self, "frame_counter", 0)
            timestamps = [1 / rate * (i + frame_counter) for i in range(len(inputs))]
            self.frame_counter = frame_counter + len(timestamps)

            # ToDo: Update the aiko fork when aiko starts putting `timestamps` in stream.variables
            #       the above code can be removed and we simply raise
            # raise ValueError("stream.variables must have 'timestamps' in order to create a DataSample object")

        data_samples = [
            DataSample(recorded_at=ts, content=i, content_type=parameters.content_type, data_file_id=uuid4())
            for i, ts in zip(inputs, timestamps)
        ]
        result = {parameters.output_key: data_samples}
        if parameters.output_empty_entities:
            result["entities"] = [{}] * len(data_samples)

        return StreamEvent.OKAY, result


class UnpackDataSamples(Capability):

    class StreamParameters(Capability.StreamParameters):
        output_key: str

    def process_frame(
        self, stream, data_samples: List[DataSample], *args, **kwargs
    ) -> Tuple[StreamEvent, Dict]:
        output = [d.content for d in data_samples]
        output_key = self.stream_parameters(stream.stream_id).output_key
        return StreamEvent.OKAY, {output_key: output}



import logging
from collections import defaultdict
from datetime import datetime, timezone
from typing import Dict, List, Optional
from uuid import UUID, uuid4

import pandas as pd
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from ....client import (
    Annotation,
    DatumSource,
    Entity,
    HLClient,
    ObjectClass,
    Observation,
)
from ....client import SubmissionType as Submission
from ....client import (
    create_assessment_from_entities,
)
from ....core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    GQLBaseModel,
    LabeledUUID,
)
from ...interfaces import IWriter

logger = logging.getLogger(__name__)


def is_valid_uuid(val):
    if isinstance(val, (LabeledUUID, UUID)):
        return True
    try:
        UUID(val)
        return True
    except (ValueError, TypeError):
        return False


class CreateSubmissionPayload(GQLBaseModel):
    submission: Optional[Submission] = None
    errors: List[str]


class HighlighterAssessmentsWriter(IWriter):
    format_name = "highlighter_assessments"

    def __init__(
        self,
        client: HLClient,
        workflow_id: int,
        object_class_uuid_lookup: Optional[Dict[str, str]] = None,
        user_id: Optional[int] = None,
        task_lookup: Optional[Dict[str, str]] = None,
    ):
        self.client = client
        self.user_id = user_id
        self.workflow_id = workflow_id
        self.task_lookup = task_lookup
        self.object_class_uuid_lookup = object_class_uuid_lookup or self._fetch_class_map()

    def _fetch_class_map(self):
        # fetch server id mapping for labels
        class Workflow(GQLBaseModel):
            object_classes: List[ObjectClass]

        ocs = self.client.workflow(return_type=Workflow, id=self.workflow_id).object_classes
        lookup = {}
        for o in ocs:
            lookup[o.name.lower()] = o.uuid
            lookup[o.name] = o.uuid
            lookup[str(o.id)] = o.uuid
        return lookup

    def _to_uuid(self, name):
        # turn a label name into a server uuid
        if isinstance(name, str):
            return str(
                self.object_class_uuid_lookup.get(name.lower(), self.object_class_uuid_lookup.get(name, name))
            )
        return name

    def _create_assessment_from_file(self, file_id, file_uuid, grp):
        # loop through all entities associated with a single file
        entities = {}
        iterator = [] if grp.empty else grp.groupby("entity_id")
        for eid, ent_df in iterator:
            attrs = ent_df.to_dict("records")
            # TODO: exception handling for entity having multiple geometry attributes
            pixel_loc = next(
                (a for a in attrs if str(a["attribute_id"]) == str(PIXEL_LOCATION_ATTRIBUTE_UUID)), None
            )
            obj_class = next(
                (a for a in attrs if str(a["attribute_id"]) == str(OBJECT_CLASS_ATTRIBUTE_UUID)), None
            )

            # seed UUID with existing entity ID if possible, random otherwise
            entity = Entity(id=UUID(str(eid)) if is_valid_uuid(eid) else uuid4())

            # if this entity had geometry (pixel_loc) and a label (obj_class),
            # we should represent it to Highlighter as an Annotation
            if pixel_loc and obj_class and file_uuid:
                # package a geometric shape and its label for the server
                geom = pixel_loc["value"]
                loc = geom.value if hasattr(geom, "value") and isinstance(geom.value, BaseGeometry) else geom
                anno = Annotation(
                    location=loc,
                    datum_source=DatumSource(confidence=float(pixel_loc["confidence"]), frame_id=0),
                    data_file_id=file_uuid,
                    entity=entity,
                    id=entity.id if is_valid_uuid(eid) else uuid4(),
                )
                # write this entity's class name attribute to an observation,
                # linked back to its data structure in Highlighter (the anno.)
                Observation(
                    attribute_id=OBJECT_CLASS_ATTRIBUTE_UUID,
                    value=self._to_uuid(obj_class["value"]),
                    confidence=float(obj_class["confidence"]),
                    annotation=anno,
                )
                # we should ignore this attribute
                ignored = {id(pixel_loc), id(obj_class)}
                parent_kw = {"annotation": anno}
            else:
                # handle simple data points without shape
                ignored = set()
                parent_kw = {"entity": entity}

            for a in attrs:
                # if we have previously parsed this attribute, or its UUID is
                # invalid
                if id(a) in ignored or not is_valid_uuid(a["attribute_id"]):
                    continue

                val = self._to_uuid(a["value"])
                if hasattr(val, "wkt"):
                    val = val.wkt
                elif isinstance(val, BaseGeometry):
                    val = val.wkt

                # package attribute as observation and inform Highlighter of its
                # structure (entity or annotation) via parent_kw
                Observation(
                    attribute_id=a["attribute_id"],
                    value=val,
                    datum_source=DatumSource(confidence=float(a["confidence"])),
                    **parent_kw,
                )
            entities[entity.id] = entity

        task_id = self.task_lookup.get(str(file_id)) if self.task_lookup else None
        try:
            task_id = UUID(str(task_id)) if task_id else None
        except (ValueError, TypeError):
            task_id = None

        submission = create_assessment_from_entities(
            client=self.client,
            entities=entities,
            data_file_ids=[file_uuid or file_id],
            task_id=task_id,
            workflow_id=self.workflow_id,
            user_id=self.user_id,
            started_at=datetime.now(timezone.utc),
        )
        return file_id, submission.id, submission.hash_signature

    def write(self, dataset):
        # pre-calculate file uuid map
        uuid_map = {}
        if "data_file_uuid" in dataset.data_files_df.columns:
            # ensure string keys for robust lookup
            df = dataset.data_files_df
            uuid_map = dict(zip(df["data_file_id"].astype(str), df["data_file_uuid"]))

        summary = []
        # group this dataset's annotations by file ID
        groups = dataset.annotations_df.groupby("data_file_id") if not dataset.annotations_df.empty else None
        # iterate over every unique file ID in the dataframe
        for fid in tqdm(dataset.data_files_df["data_file_id"].unique()):
            # extract the subset of all annotations associated with this file ID,
            # and upload them
            grp = groups.get_group(fid) if groups and fid in groups.groups else pd.DataFrame()
            file_uuid = uuid_map.get(str(fid))
            summary.append(self._create_assessment_from_file(fid, file_uuid, grp))

        # link server identity back to the local dataset
        id_map = {s[0]: s[1] for s in summary}
        sig_map = {s[0]: s[2] for s in summary}

        dataset.data_files_df["assessment_id"] = dataset.data_files_df["data_file_id"].map(id_map)
        dataset.data_files_df["hash_signature"] = dataset.data_files_df["data_file_id"].map(sig_map)



import errno as errno_module
import re
from typing import Any, Dict, Optional


def format_ffmpeg_error_message(operation_name: str, error: Exception, source: object) -> str:
    details = extract_ffmpeg_error_details(error)
    reason = details.get("strerror") or details.get("message") or "Unknown FFmpeg error"
    if "last error log:" in reason:
        reason = reason.split("last error log:", 1)[0].strip().rstrip(";")

    parts = [f"RETRY: {operation_name} failed with FFmpeg error: {reason}"]
    errno_val = details.get("errno")
    if errno_val is not None:
        errno_name = details.get("errno_name")
        if errno_name:
            parts.append(f"errno={errno_val}({errno_name})")
        else:
            parts.append(f"errno={errno_val}")
    if details.get("ffmpeg_log"):
        parts.append(f"ffmpeg_log={details['ffmpeg_log']}")
    if details.get("hint"):
        parts.append(f"hint={details['hint']}")
    parts.append(f"Source: {source}")
    return " | ".join(parts)


def extract_ffmpeg_error_details(error: Exception) -> Dict[str, Any]:
    message = str(error)
    details: Dict[str, Any] = {"message": message}
    errno_val = getattr(error, "errno", None)
    if errno_val is not None:
        details["errno"] = errno_val
        details["errno_name"] = errno_module.errorcode.get(errno_val)
        details["strerror"] = getattr(error, "strerror", None)

    last_log = None
    error_log = getattr(error, "log", None)
    if error_log:
        try:
            last_log = error_log[-1]
        except Exception:
            last_log = None
        if last_log is not None and not isinstance(last_log, str):
            last_log = str(last_log)
    if not last_log:
        match = re.search(r"last error log:\s*(.+)$", message)
        if match:
            last_log = match.group(1).strip()
    if last_log:
        details["ffmpeg_log"] = last_log

    details["hint"] = ffmpeg_error_hint(
        message=message,
        last_log=details.get("ffmpeg_log"),
        errno_name=details.get("errno_name"),
        errno_val=details.get("errno"),
    )
    return details


def ffmpeg_error_hint(
    message: str,
    last_log: Optional[str],
    errno_name: Optional[str],
    errno_val: Optional[int],
) -> Optional[str]:
    combined = f"{message} {last_log or ''}".lower()

    if "time_scale/num_units_in_tick" in combined:
        return "H.264 timing metadata is invalid; check camera/encoder firmware or stream settings."
    if "invalid data found when processing input" in combined:
        return "Stream data looks corrupted or codec parameters mismatch; check encoder output."
    if "connection reset" in combined or errno_name in {"ECONNRESET", "ECONNREFUSED", "ETIMEDOUT"}:
        return "Network interruption from source; check camera/network stability."
    if errno_name in {"EIO", "ENETDOWN", "ENETUNREACH", "EHOSTUNREACH"} or errno_val in {
        errno_module.EIO,
        errno_module.ENETDOWN,
        errno_module.ENETUNREACH,
        errno_module.EHOSTUNREACH,
    }:
        return "I/O error while reading the stream; often network or camera instability."
    if "end of file" in combined:
        return "Source ended or closed the stream."

    return None



from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from highlighter.core.data_models import DataSample
from highlighter.core.exceptions import require_package

try:
    import cv2
except ModuleNotFoundError as _:
    cv2 = None
import numpy as np
from PIL import Image
from pydantic import BaseModel, field_validator
from shapely import affinity
from shapely import geometry as gm

PathLike = Union[str, Path]
LTRBRect = Tuple[int, int, int, int]

KEY_CROP_ROTATED_RECT = "crop_rotated_rect"

__all__ = [
    "CropArgs",
    "KEY_CROP_ROTATED_RECT",
    "crop_rect_from_poly",
    "crop_rectangle",
    "crop_rotated_rectangle",
]


class CropArgs(BaseModel):
    crop_rotated_rect: Optional[bool] = False
    scale: Optional[float] = None
    pad: Optional[int] = None
    warped_wh: Optional[Union[Tuple[int, int], List[int]]] = None

    @field_validator("warped_wh")
    @classmethod
    def _validate_warped_wh(cls, v):
        """If warped_wh is not None make sure it's a list and not a tuple
        because tuple is not json/safe_yaml serializable.
        """

        if isinstance(v, tuple):
            v = list(v)

        if v is not None:
            assert isinstance(v, list)
            assert len(v) == 2
        return v

    def dict(self, **kwargs):
        d = super().model_dump(**kwargs)
        if not self.crop_rotated_rect:
            d.pop("warped_wh")
        return d

    def __call__(
        self, crop_locations: List[LTRBRect | gm.Polygon], image: Union[DataSample, np.ndarray, Image.Image]
    ):
        if isinstance(image, DataSample):
            return image.crop_content(crop_locations, crop_args=self, as_data_sample=True)
        else:
            if self.crop_rotated_rect:
                crop_fn = crop_rotated_rectangle
            else:
                crop_fn = crop_rectangle

            kwargs = self.model_dump()

            polys = []
            for loc in crop_locations:
                if hasattr(loc, "bounds"):
                    polys.append(loc)
                elif isinstance(loc, (list, tuple)) and len(loc) == 4:
                    x0, y0, x1, y1 = loc
                    polys.append(gm.Polygon([(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)]))
                else:
                    raise ValueError(f"Invalid crop location, got {loc}")

            return [crop_fn(image, poly, **kwargs) for poly in polys]


def _get_wh_of_rotated_rect(np_corners: np.ndarray) -> Tuple[int, int]:
    # First 3 corners of the rectangle
    a, b, c = np_corners[0], np_corners[1], np_corners[2]
    ab = int(round(np.linalg.norm(b - a)))
    bc = int(round(np.linalg.norm(c - b)))
    if ab >= bc:
        w, h = ab, bc
    else:
        w, h = bc, ab

    return w, h


def _order_rect_points_bottom_left_ccw(rect: gm.Polygon) -> gm.Polygon:
    # 0. Note, the origin for Shapely is the bottom left with
    #    the positive x and y axis right and up respectivally.
    #    But, our coordinates have the origin at the top left
    #    and positive x,y being rigth and down.
    #    So our top right and clockwise will translate to shapely's
    #    Bottom left and counter-clockwise. See below for an illustration
    #
    #    Shapely coords:
    #
    #    +y
    #    |
    #    | 0,1      2,1
    #    * * * * * *
    #    *         *
    #    * * * * * * --- +x
    #     0,0      2,0

    #    Highlighter coords:
    #
    #     0,0      2,0
    #    * * * * * * --- +x
    #    *         *
    #    * * * * * *
    #    |0,1      2,1
    #    |
    #    +y
    #
    #
    # The following uses the shapely coordinate system, that way, if we
    # look for bottom left and order ccw we should be golden
    #
    # 1. Find the edges that belong to the major axis (longest)
    # 2. Select the bottom edge
    # 3. Find the left most point of the bottom edge
    # 4. Set bottom left point to 0th index
    # 5. remaining points follow ccw

    rect_points = np.array(rect.exterior.coords)[:-1]

    # Find major axis
    a, b, c, d = 0, 1, 2, 3
    x, y = 0, 1
    ab = gm.LineString((rect_points[a], rect_points[b]))
    bc = gm.LineString((rect_points[b], rect_points[c]))
    cd = gm.LineString((rect_points[c], rect_points[d]))
    da = gm.LineString((rect_points[d], rect_points[a]))
    edges = np.array([ab, bc, cd, da])
    longest = edges[np.argsort([edge.length for edge in edges])[-2:]]

    edge_0 = np.array(longest[0].coords)
    edge_1 = np.array(longest[1].coords)

    # Find bottom edge
    if edge_0[..., y].min() < edge_1[..., y].min():
        bottom_edge = edge_0
        top_edge = edge_1
    else:
        bottom_edge = edge_1
        top_edge = edge_0

    # arrange points starting from the left most point of the bottom
    # edge and moving counter clockwise
    if bottom_edge[0][x] > bottom_edge[1][x]:
        bottom_edge = bottom_edge[[1, 0]]
        top_edge = top_edge[[1, 0]]

    return gm.Polygon(bottom_edge.tolist() + top_edge.tolist())


def crop_rect_from_poly(
    image: Union[np.ndarray, Image.Image],
    polys: Union[gm.Polygon, List[gm.Polygon]],
    crop_args: Optional[Union[CropArgs, Dict]] = None,
) -> Union[List[np.ndarray], List[Image.Image], np.ndarray, Image.Image]:
    single_crop = False
    if isinstance(polys, gm.Polygon):
        polys = [polys]
        single_crop = True

    if crop_args is None:
        crop_args = {}

    if isinstance(crop_args, dict):
        crop_args = CropArgs(**crop_args)
    elif not isinstance(crop_args, CropArgs):
        raise ValueError(f"invalid parameter 'crop_args' expected a CropArgs object, got {crop_args}")

    kwargs = crop_args.dict()

    crop_rotated_rect = kwargs.pop(KEY_CROP_ROTATED_RECT)

    if crop_rotated_rect:
        crop_fn = crop_rotated_rectangle
    else:
        crop_fn = crop_rectangle

    crops = [crop_fn(image, poly, **kwargs) for poly in polys]
    if single_crop:
        return crops[0]
    else:
        return crops


@require_package(cv2, "cv2", "opencv")
def crop_rotated_rectangle(
    image: Union[np.ndarray, Image.Image],
    poly: gm.Polygon,
    scale: Optional[float] = None,
    pad: Optional[int] = None,
    warped_wh: Optional[Tuple[int, int]] = None,
) -> Union[np.ndarray, Image.Image]:
    """Find the minimum_rotated_rectange for the input poly and apply
    cv2.warpPerspective to transform to to rectangular crop of shape warped_wh

    see: https://shapely.readthedocs.io/en/stable/manual.html#object.minimum_rotated_rectangle

    Resulting crops will be of the same type as the input image (np.ndarray | PIL.Image)

    Args:
        image: Image to crop
        poly: Polygon to crop out
        scale: (Optional) factor applied to poly before cropping about the origin
        pad: Optional pad (in pixels) applied to poly before cropping
        warped_wh: The desired output shape of the crop
    """
    assert isinstance(pad, int) or (
        pad is None
    ), f"Croping rotated rec only support a int for 'pad', got: {pad}"

    rot_rect = poly.minimum_rotated_rectangle

    if scale is not None:
        rot_rect = affinity.scale(rot_rect, xfact=scale, yfact=scale, origin=(0, 0, 0))

    if pad is not None:
        rot_rect = rot_rect.buffer(pad, join_style=2)

    # Shapely uses a different coordinate system to us and opencv.
    # We need to make sure the points in the src_points corrispond to
    # the points in dst_points, hence this function is needed. More detail in
    # its doc string.
    rot_rect = _order_rect_points_bottom_left_ccw(rot_rect)
    src_points = np.array(list(zip(*rot_rect.exterior.coords.xy)))[:-1].astype(np.float32)

    if warped_wh is None:
        warped_wh = _get_wh_of_rotated_rect(src_points)
    w, h = warped_wh

    dst_points = np.array([[0, 0], [w, 0], [w, h], [0, h]], dtype=np.float32)

    transform_mat = cv2.getPerspectiveTransform(src_points, dst_points)

    if isinstance(image, Image.Image):
        src = np.array(image)
    elif isinstance(image, np.ndarray):
        src = image
    else:
        raise ValueError(f"Invalid image input type: {type(image)}")

    out = cv2.warpPerspective(src, transform_mat, warped_wh, flags=cv2.INTER_LINEAR)

    return out if isinstance(image, np.ndarray) else Image.fromarray(out)


def crop_rectangle(
    image: Union[np.ndarray, Image.Image],
    rect: gm.Polygon,
    scale: Optional[float] = None,
    pad: Optional[int] = None,
) -> Union[np.ndarray, Image.Image]:
    """Crop an XY Axis aligned rectangle crop image

    Resulting crops will be of the same type as the input image (np.ndarray | PIL.Image)

    Args:
        image: Image to crop
        rect: Left, Top, Right, Bottom tuple representing the rectangle to crop
        scale: Optional factor applied to rect before cropping
        pad: Optional pad (in pixels) applied to rect before cropping

    Note: Pillow.Image.crop does not give a 💩 if a crop goes out-of-bounds
    it will just add zeros as needed.
    """
    if hasattr(rect, "bounds"):
        # shapely object
        rect = rect.bounds
    else:
        raise ValueError("rect argument must have a .bounds attribute in order to crop")

    if scale is not None:
        rect = [x * scale for x in rect]

    if pad is None:
        pad = (0, 0)
    elif isinstance(pad, (float, int)):
        pad = (int(pad), int(pad))
    assert isinstance(pad, tuple)

    def pad_rect(rect, w_pad, h_pad):
        l, t, r, b = rect
        l -= w_pad
        r += w_pad
        t -= h_pad
        b += h_pad
        return (l, t, r, b)

    l, t, r, b = [int(i) for i in pad_rect(rect, *pad)]

    if isinstance(image, Image.Image):
        crop = image.crop((l, t, r, b))
    elif isinstance(image, np.ndarray):
        crop = np_crop_with_auto_pad(image, (l, t, r, b))
    else:
        raise ValueError(f"Invalid image input type: {type(image)}")
    return crop


def np_crop_with_auto_pad(
    img: np.ndarray,
    xyxybox: Tuple[int, int, int, int],
) -> np.ndarray:
    """
    Crops an image using NumPy. If the crop region is partially or fully
    outside the image bounds, pads with zeros. This matches the behaviour
    of PIL.Image.crop

    Parameters:
    -----------
    img : np.ndarray
        Input image (H, W) or (H, W, C)
    y, x : int
        Top-left corner of the crop (can be negative)
    h, w : int
        Desired height and width of output crop

    Returns:
    --------
    np.ndarray
        Cropped (and possibly zero-padded) image of shape (h, w) or (h, w, C)
    """
    x0, y0, x1, y1 = xyxybox
    h = y1 - y0
    w = x1 - x0
    if h <= 0 or w <= 0:
        raise ValueError("Crop height and width must be positive")

    # Get image dimensions
    img_h, img_w = img.shape[:2]
    channels = img.shape[2] if img.ndim == 3 else 1

    # Create output array filled with zeros
    dtype = img.dtype
    if img.ndim == 3:
        out = np.zeros((h, w, channels), dtype=dtype)
    else:
        out = np.zeros((h, w), dtype=dtype)

    # Calculate source and destination regions
    # src_y_start, src_y_end → region in original image
    # dst_y_start, dst_y_end → region in output image

    src_y_start = max(y0, 0)
    src_y_end = min(y1, img_h)
    src_x_start = max(x0, 0)
    src_x_end = min(x1, img_w)

    dst_y_start = max(-y0, 0)
    dst_y_end = dst_y_start + (src_y_end - src_y_start)
    dst_x_start = max(-x0, 0)
    dst_x_end = dst_x_start + (src_x_end - src_x_start)

    # If there's any overlap → copy the valid region
    if (src_y_start < src_y_end) and (src_x_start < src_x_end):
        out[dst_y_start:dst_y_end, dst_x_start:dst_x_end] = img[src_y_start:src_y_end, src_x_start:src_x_end]

    return out



# To Do
# ~~~~~
# - None, yet.
#
# Notes
# ~~~~~
# gst-launch-1.0 filesrc location=../football_test_1.mp4 ! qtdemux ! avdec_h264 ! videoconvert ! osxvideosink
#
# CUDA_VISIBLE_DEVICES=3 /mnt/code/aiko_services_internal/aiko_services_internal/video/gstreamer/video_example.py -if /app/resources/sample.mp4 -os 192.168.1.154:5000 -r 1280 1280 -f 6/1

import os

from aiko_services.elements.gstreamer import *

__all__ = ["VideoFileReader"]

# -----------------------------------------------------------------------------

class VideoFileReader:
  def __init__(self, input_filename, width, height):
    if not os.path.exists(input_filename):
      raise ValueError("File does not exist: " + input_filename)

    Gst = gst_initialise()

    gst_launch_command = "filesrc location={} ! qtdemux ! {} ! videoconvert ! video/x-raw, format={} ! appsink name=sink".format(input_filename, utilities.get_h264_decoder(), utilities.get_format())
    pipeline = Gst.parse_launch(gst_launch_command)

    sink = pipeline.get_by_name("sink")
    sink_caps = "video/x-raw, format={}, width={}, height={}".format(utilities.get_format(), width, height)
    sink.set_property("caps", Gst.caps_from_string(sink_caps))

    self.video_reader = VideoReader(pipeline, sink)

  def queue_size(self):
    return self.video_reader.queue.qsize()

  def read_frame(self, timeout = None):
    return self.video_reader.read_frame(timeout)



# Usage
# ~~~~~
# pytest [-s] unit/test_stream_event.py
# pytest [-s] unit/test_stream_event.py::test_stream_error
#
# test_stream_error()
# - PipelineElement.process_frame() returning StreamEvent.ERROR
#   causes Pipeline to "hang, due to Stream.lock contention.
#   See https://github.com/geekscape/aiko_services/pull/32
#
# To Do
# ~~~~~
# - Improve Aiko Services Process exit
# - Test StreamEvent.OKAY, StreamEvent.STOP, StreamEvent.ERROR for ...
#   - create_stream, _create_frame_generator, process_frame, destroy_stream
#   - Consolidate any overlap with "test_stream_lock.py" ?

from typing import Tuple

import aiko_services as aiko
from aiko_services.tests.unit import do_create_pipeline

PIPELINE_DEFINITION = """{
  "version": 0, "name": "p_test", "runtime": "python",
  "graph": ["(StreamError)"],
  "elements": [
    { "name":   "StreamError", "input":  [], "output": [],
      "deploy": {
        "local": { "module": "aiko_services.tests.unit.test_stream_event" }
      }
    }
  ]
}
"""

class StreamError(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("stream_error:0")
        context.call_init(self, "PipelineElement", context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"process_frame(): {self.my_id()}")
        aiko.process.terminate()  # TODO: Improve Aiko Services Process exit
        diagnostic = "Testing StreamEvent.ERROR"
        return aiko.StreamEvent.ERROR, {"diagnostic": diagnostic}

def test_stream_error():
    do_create_pipeline(PIPELINE_DEFINITION)



#!/usr/bin/env python3
#
# https://en.wikipedia.org/wiki/Chaos_engineering#Chaos_Monkey
#
# TCP reachability chaos monkey with modes
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Up:   Proxy frontend → backend
# Down: hardfail:       Frontend port closed (connection refused/reset)
# Down: blackhole:      Frontend accepts but never forwards (timeout)
# Down: drop_midstream: Proxy forwards but randomly kills sessions mid-flow
#
# "Backend":  TCP/IP port that the server is listening to
# "Frontend": TCP/IP port that can be controlled for chaos monkey testing
#
# Usage
# ~~~~~
# ./network_chaos_monkey.py status
# ./network_chaos_monkey.py up --backend_port 1883 --frontend_port 9883  \
#                             [--no_precheck]
#
# ./network_chaos_monkey.py down   --mode hardfail
#
# ./network_chaos_monkey.py toggle --mode hardfail --on 10 --off 10  \
#                                  --backend_port 1883 --frontend_port 9883
#
# ./network_chaos_monkey.py toggle --mode drop_midstream  \
#         --on 8 --off 5    --cycles 4 --jitter 2         \
#         --drop_chance 0.3 --drop_min_delay 2 --drop_max_delay 12
#
# ./network_chaos_monkey.py stop
#
# mosquitto_sub -h 127.0.0.1 -p 9883 -t test -V mqttv311 -v -d
#
# To Do
# ~~~~~
# - Implement as an Actor, which directly uses the MQTT backend port
# - "aiko/service/network_chaos_monkey" bootstrap in case Registar unavailable ?

import argparse
import asyncio
from contextlib import suppress
import os
import random
import signal
import socket
import subprocess
import sys
import time
from typing import Optional, Set

PIDFILE = "network_chaos_monkey.pid"
MODEFILE = "network_chaos_monkey.mode"   # "proxy", "sink", or "drop_midstream"

ACTIVE_CONNECTIONS: Set[asyncio.StreamWriter] = set()

# -------------
# Async servers
# -------------

async def _pump(src: asyncio.StreamReader, dst: asyncio.StreamWriter):
    try:
        while True:
            data = await src.read(65536)
            if not data:
                break
            dst.write(data)
            await dst.drain()
    except Exception:
        pass
    finally:
        with suppress(Exception):
            dst.close()

async def _handle_proxy_client(
    cli_reader, cli_writer, backend_host, backend_port,
    drop_midstream=False, drop_chance=0.2,
    drop_min_delay=1.0, drop_max_delay=5.0):

    peer = None
    try:
        peer = cli_writer.get_extra_info("peername")
        print(f"[proxy] client connected: {peer}", flush=True)
        be_reader, be_writer = await asyncio.open_connection(
            backend_host, backend_port)
    except Exception as e:
        # Explicit message so it's obvious in logs what happened.
        print(f"[proxy] backend connect failed for client {peer}: {backend_host}:{backend_port} -> {e}", flush=True)
        try:
            cli_writer.close()
        except Exception:
            pass
        return

    if drop_midstream:
        ACTIVE_CONNECTIONS.add(cli_writer)

        async def killer():
            mn = max(0.0, float(min(drop_min_delay, drop_max_delay)))
            mx = max(0.0, float(max(drop_min_delay, drop_max_delay)))
            delay = random.uniform(mn, mx)
            await asyncio.sleep(delay)
            if random.random() < max(0.0, min(1.0, float(drop_chance))) and not cli_writer.is_closing():
                print(f"[drop_midstream] killing session {peer} after {delay:.2f}s", flush=True)
                try:
                    cli_writer.close()
                except Exception:
                    pass
                try:
                    be_writer.close()
                except Exception:
                    pass

        asyncio.create_task(killer())

    t1 = asyncio.create_task(_pump(cli_reader, be_writer))   # client -> backend
    t2 = asyncio.create_task(_pump(be_reader, cli_writer))   # backend -> client
    await asyncio.gather(t1, t2)

    if drop_midstream:
        ACTIVE_CONNECTIONS.discard(cli_writer)

async def _run_proxy(
    front_host, front_port, backend_host, backend_port,
    drop_midstream=False, drop_chance=0.2,
    drop_min_delay=1.0, drop_max_delay=5.0):

    server = await asyncio.start_server(
        lambda r, w: _handle_proxy_client(
            r, w, backend_host, backend_port,
            drop_midstream, drop_chance, drop_min_delay, drop_max_delay),
            front_host, front_port
    )
    tag = "[proxy+drop_midstream]" if drop_midstream else "[proxy]"
    print(f"{tag} UP: {front_host}:{front_port} → {backend_host}:{backend_port}", flush=True)
    async with server:
        await server.serve_forever()

async def _handle_sink_client(cli_reader, cli_writer):
    try:
        while True:
            data = await cli_reader.read(65536)
            if not data:
                break
            # discard; never respond
    except Exception:
        pass
    finally:
        try:
            cli_writer.close()
        except Exception:
            pass

async def _run_sink(front_host, front_port):
    server = await asyncio.start_server(
        _handle_sink_client, front_host, front_port)
    print(f"[sink] DOWN (blackhole): {front_host}:{front_port}", flush=True)
    async with server:
        await server.serve_forever()

# ------------
# Daemon entry
# ------------

def _serve_entry(args):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        if args.behavior == "proxy":
            coro = _run_proxy(args.host, args.frontend_port,
                              args.backend_host, args.backend_port,
                              drop_midstream=False)
        elif args.behavior == "sink":
            coro = _run_sink(args.host, args.frontend_port)
        elif args.behavior == "drop_midstream":
            coro = _run_proxy(args.host, args.frontend_port,
                              args.backend_host, args.backend_port,
                              drop_midstream=True,
                              drop_chance=args.drop_chance,
                              drop_min_delay=args.drop_min_delay,
                              drop_max_delay=args.drop_max_delay)
        else:
            print(f"Unknown behavior: {args.behavior}", file=sys.stderr)
            sys.exit(2)
        loop.run_until_complete(coro)
    except KeyboardInterrupt:
        pass
    finally:
        try:
            loop.close()
        except Exception:
            pass

# ---------------
# Process helpers
# ---------------

def _read_pid() -> Optional[int]:
    try:
        with open(PIDFILE, "r") as f:
            return int(f.read().strip())
    except Exception:
        return None

def _write_pid(pid: int):
    with open(PIDFILE, "w") as f:
        f.write(str(pid))

def _write_modefile(behavior: str):
    with open(MODEFILE, "w") as f:
        f.write(behavior)

def _read_modefile() -> Optional[str]:
    try:
        with open(MODEFILE, "r") as f:
            return f.read().strip()
    except Exception:
        return None

def _is_running(pid: Optional[int]) -> bool:
    if pid is None:
        return False
    try:
        os.kill(pid, 0)
        return True
    except Exception:
        return False

def _kill_pid(pid: Optional[int]):
    if pid is None:
        return
    try:
        os.kill(pid, signal.SIGTERM)
    except Exception:
        if sys.platform.startswith("win"):
            try:
                os.kill(pid, signal.SIGBREAK)
            except Exception:
                pass

def _spawn_detached(argv_for_child) -> int:
    kwargs = {}
    if os.name == "posix":
        kwargs["start_new_session"] = True
    elif os.name == "nt":
        DETACHED_PROCESS = 0x00000008
        CREATE_NEW_PROCESS_GROUP = 0x00000200
        kwargs["creationflags"] = DETACHED_PROCESS | CREATE_NEW_PROCESS_GROUP
    proc = subprocess.Popen(argv_for_child,
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.DEVNULL,
                            stdin=subprocess.DEVNULL,
                            **kwargs)
    return proc.pid

def _precheck_backend(host: str, port: int, timeout: float = 1.5) -> bool:
    """Synchronous reachability check to avoid 'UP' when backend is down."""
    try:
        with socket.create_connection((host, port), timeout=timeout):
            return True
    except Exception as e:
        print(f"[precheck] backend {host}:{port} unreachable: {e}", file=sys.stderr)
        return False

# --------
# Commands
# --------

def cmd_up(args):
    pid = _read_pid()
    if _is_running(pid):
        print(f"Already UP (PID {pid})")
        return

    behavior = "proxy"
    if args.mode == "drop_midstream":
        behavior = "drop_midstream"

    if args.precheck and not _precheck_backend(args.backend_host, args.backend_port, args.precheck_timeout):
        # Fail fast so the mapping is truly symmetric & predictable.
        sys.exit(1)

    serve_argv = [sys.executable, __file__, "serve",
                  "--behavior", behavior,
                  "--host", args.host,
                  "--frontend_port", str(args.frontend_port),
                  "--backend_host", args.backend_host,
                  "--backend_port", str(args.backend_port),
                  "--drop_chance", str(args.drop_chance),
                  "--drop_min_delay", str(args.drop_min_delay),
                  "--drop_max_delay", str(args.drop_max_delay)]
    child_pid = _spawn_detached(serve_argv)
    _write_pid(child_pid)
    _write_modefile(behavior)
    print(f"UP: {behavior} {args.host}:{args.frontend_port} → {args.backend_host}:{args.backend_port} (PID {child_pid})")

def cmd_down(args):
    mode = args.mode
    pid = _read_pid()

    if mode == "hardfail":
        if _is_running(pid):
            _kill_pid(pid)
            time.sleep(0.2)
        for f in (PIDFILE, MODEFILE):
            if os.path.exists(f):
                try:
                    os.remove(f)
                except Exception:
                    pass
        print("DOWN: hardfail (port closed).")

    elif mode == "blackhole":
        if _is_running(pid):
            _kill_pid(pid)
            time.sleep(0.2)
        serve_argv = [sys.executable, __file__, "serve",
                      "--behavior", "sink",
                      "--host", args.host,
                      "--frontend_port", str(args.frontend_port)]
        child_pid = _spawn_detached(serve_argv)
        _write_pid(child_pid)
        _write_modefile("sink")
        print(f"DOWN: blackhole at {args.host}:{args.frontend_port} (PID {child_pid})")

    elif mode == "drop_midstream":
        if _is_running(pid):
            _kill_pid(pid)
            time.sleep(0.2)
        serve_argv = [sys.executable, __file__, "serve",
                      "--behavior", "drop_midstream",
                      "--host", args.host,
                      "--frontend_port", str(args.frontend_port),
                      "--backend_host", args.backend_host,
                      "--backend_port", str(args.backend_port),
                      "--drop_chance", str(args.drop_chance),
                      "--drop_min_delay", str(args.drop_min_delay),
                      "--drop_max_delay", str(args.drop_max_delay)]
        child_pid = _spawn_detached(serve_argv)
        _write_pid(child_pid)
        _write_modefile("drop_midstream")
        print(f"DOWN: drop_midstream at {args.host}:{args.frontend_port} (PID {child_pid})")

def cmd_status(_args):
    pid = _read_pid()
    if _is_running(pid):
        beh = _read_modefile() or "unknown"
        print(f"Status: running ({beh}) PID {pid}")
    else:
        print("Status: DOWN (no process)")

def cmd_stop(_args):
    pid = _read_pid()
    if not _is_running(pid):
        print("Nothing to stop (no running process).")
        for f in (PIDFILE, MODEFILE):
            if os.path.exists(f):
                try:
                    os.remove(f)
                except Exception:
                    pass
        return

    print(f"Stopping chaos monkey (PID {pid}) ...")
    _kill_pid(pid)
    time.sleep(0.2)

    for f in (PIDFILE, MODEFILE):
        if os.path.exists(f):
            try:
                os.remove(f)
            except Exception:
                pass
    print("Stopped.")

def _jittered(base: float, jitter: float) -> float:
    if jitter <= 0:
        return max(base, 0.1)
    delta = random.uniform(-jitter, +jitter)
    return max(base + delta, 0.1)

def cmd_toggle(args):
    if args.seed is not None:
        random.seed(args.seed)

    j_on  = args.jitter_on  if args.jitter_on  is not None else args.jitter
    j_off = args.jitter_off if args.jitter_off is not None else args.jitter

    for i in range(1, args.cycles + 1):
        on_dur  = _jittered(args.on,  j_on)
        off_dur = _jittered(args.off, j_off)

        print(f"[cycle {i}] UP for {on_dur:.2f}s")
        cmd_up(args)
        time.sleep(on_dur)

        print(f"[cycle {i}] DOWN for {off_dur:.2f}s (mode={args.mode})")
        cmd_down(args)
        time.sleep(off_dur)

# ----------
# CLI parser
# ----------

def parse_args():
    p = argparse.ArgumentParser(description="Network Chaos Monkey (proxy + hardfail/blackhole/drop_midstream)")
    sub = p.add_subparsers(dest="cmd", required=True)

    common = argparse.ArgumentParser(add_help=False)
    common.add_argument("--host", default="0.0.0.0", help="Frontend bind address")
    common.add_argument("--frontend_port", type=int, default=1883, help="Frontend port for tcp_client connections")
    common.add_argument("--backend_host", default="127.0.0.1", help="Backend server host (when proxying)")
    common.add_argument("--backend_port", type=int, default=9883, help="Backend server port (when proxying)")
    # drop_midstream tuning (safe to pass always; ignored unless behavior == drop_midstream)
    common.add_argument("--drop_chance", type=float, default=0.2, help="Probability [0..1] to kill a session")
    common.add_argument("--drop_min_delay", type=float, default=1.0, help="Earliest kill (seconds)")
    common.add_argument("--drop_max_delay", type=float, default=5.0, help="Latest kill (seconds)")
    # precheck toggle
    common.add_argument("--no_precheck", dest="precheck", action="store_false",
                        help="Do not verify backend reachability before UP")
    common.add_argument("--precheck", dest="precheck", action="store_true",
                        help=argparse.SUPPRESS)
    common.add_argument("--precheck_timeout", type=float, default=1.5, help="Backend precheck timeout (seconds)")
    common.set_defaults(precheck=True)

    up = sub.add_parser("up", parents=[common], help="Start UP state (proxy or drop_midstream)")
    up.add_argument("--mode", choices=["hardfail", "blackhole", "drop_midstream"], default="hardfail",
                    help="UP behavior: proxy (hardfail), or proxy with drop_midstream")
    up.set_defaults(func=cmd_up)

    down = sub.add_parser("down", parents=[common], help="Enter DOWN state")
    down.add_argument("--mode", choices=["hardfail", "blackhole", "drop_midstream"], required=True,
                      help="DOWN behavior")
    down.set_defaults(func=cmd_down)

    status = sub.add_parser("status", help="Show current status")
    status.set_defaults(func=cmd_status)

    stop = sub.add_parser("stop", help="Stop any running chaos monkey process")
    stop.set_defaults(func=cmd_stop)

    toggle = sub.add_parser("toggle", parents=[common], help="Alternate UP/DOWN cycles")
    toggle.add_argument("--mode", choices=["hardfail", "blackhole", "drop_midstream"], required=True,
                        help="DOWN behavior during toggling")
    toggle.add_argument("--on", type=float, default=5.0, help="Seconds UP")
    toggle.add_argument("--off", type=float, default=5.0, help="Seconds DOWN")
    toggle.add_argument("--cycles", type=int, default=5, help="Number of cycles")
    toggle.add_argument("--jitter", type=float, default=0.0, help="±seconds jitter applied to both on/off")
    toggle.add_argument("--jitter_on", type=float, default=None, help="Override jitter for UP")
    toggle.add_argument("--jitter_off", type=float, default=None, help="Override jitter for DOWN")
    toggle.add_argument("--seed", type=int, default=None, help="Random seed for reproducibility")
    toggle.set_defaults(func=cmd_toggle)

    serve = sub.add_parser("serve", help=argparse.SUPPRESS)
    serve.add_argument("--behavior", choices=["proxy", "sink", "drop_midstream"], required=True)
    serve.add_argument("--host", required=True)
    serve.add_argument("--frontend_port", type=int, required=True)
    serve.add_argument("--backend_host", default="127.0.0.1")
    serve.add_argument("--backend_port", type=int, default=9883)
    serve.add_argument("--drop_chance", type=float, default=0.2)
    serve.add_argument("--drop_min_delay", type=float, default=1.0)
    serve.add_argument("--drop_max_delay", type=float, default=5.0)
    serve.set_defaults(func=lambda a: _serve_entry(a))

    return p.parse_args()

def main():
    args = parse_args()
    args.func(args)

if __name__ == "__main__":
    main()



import logging
from datetime import datetime, timezone
from typing import List
from urllib.parse import urlparse

from ..core import PIXEL_LOCATION_ATTRIBUTE_UUID, GQLBaseModel
from .gql_client import HLClient

__all__ = ["create_capability_for_artefact"]

logger = logging.getLogger(__name__)


def create_capability_for_artefact(
    client: HLClient, artefact_id: str, training_run_id: int, inference_config: dict
):
    from .training_config import TrainingConfigType, get_training_config

    # get capability schema via GraphQL
    training_config: TrainingConfigType = get_training_config(client, training_run_id)

    class TrainingRun(GQLBaseModel):
        name: str

    training_run_name = client.trainingRun(return_type=TrainingRun, id=training_run_id).name

    class CurrentUser(GQLBaseModel):
        id: int

    current_user_id = client.currentUser(return_type=CurrentUser).id

    if "parameters" in inference_config:
        parameters = inference_config["parameters"].copy()
    elif "frame_parameters" in inference_config:
        parameters = inference_config["frame_parameters"].copy()
    else:
        raise KeyError(
            f"'parameters' or 'frame_parameters' cannot be found in inference_config: {inference_config}"
        )

    # pull input filter enum set out of capability schema and put in parameters
    # as this is not currently included in the taxonomy datatypes for inputs and outputs
    if len(training_config.model_schema.model_inputs.filters) > 1:
        raise NotImplementedError("Can't create capabilities from training runs with multiple input filters")
    if len(training_config.model_schema.model_inputs.filters[0]) > 0:
        parameters["enum_array_filter_include_ids"] = (
            training_config.model_schema.get_input_filter_attribute_values()[0]
        )
        parameters["enum_array_filter_attribute"] = (
            training_config.model_schema.get_input_filter_attribute_ids()[0][0]
        )
    # pull confidences out of model schema and put in parameters
    if len(training_config.model_schema.get_head_model_outputs(1)) > 0:
        raise NotImplementedError("Can't create capability from multi-headed model")
    parameters["output_confidence_thresholds"] = {
        str(output.entity_attribute_enum_id): output.conf_thresh
        for output in training_config.model_schema.model_outputs
    }
    machine_agent_version = {
        "frameParameters": parameters,
        "streamParameters": {},
        "deploymentParameters": {"local": {"module": f"hl_train.elements"}},
        "machineAgentTypeId": inference_config["machine_agent_type_id"],
        "trainingRunId": training_run_id,
        "trainingRunArtefactId": artefact_id,
        "title": training_run_name,
        "code": inference_config["code"],
        "codeVersion": "1",
        "publishedById": current_user_id,
        "publishedAt": datetime.now(timezone.utc).isoformat(),
    }

    # define inputs and outputs
    inputs = ["data_samples"]
    for input in training_config.model_schema.model_inputs.entity_attributes:
        if input.entity_attribute_id == str(PIXEL_LOCATION_ATTRIBUTE_UUID):
            continue
        else:
            inputs.append("entities")
            break
    outputs = ["entities"]

    # run mutations
    class MachineAgentVersion(GQLBaseModel):
        id: str

    class CreateMachineAgentVersionPayload(GQLBaseModel):
        errors: List[str]
        machine_agent_version: MachineAgentVersion

    create_machine_agent_version_result = client.createMachineAgentVersion(
        return_type=CreateMachineAgentVersionPayload,
        **machine_agent_version,
    )
    if len(create_machine_agent_version_result.errors) > 0:
        raise ValueError(
            f"Could not create machine_agent_version: {create_machine_agent_version_result.errors}"
        )
    machine_agent_version_id = create_machine_agent_version_result.machine_agent_version.id

    class CreatePipelineElementInputPayload(GQLBaseModel):
        errors: List[str]

    for i, input in enumerate(inputs):
        create_input_result = client.createPipelineElementInput(
            return_type=CreatePipelineElementInputPayload,
            machineAgentVersionId=machine_agent_version_id,
            name=input,
            arrayNestingLevels=0,
            sortOrder=str(i),
            isAnnotation=True,
        )
        if len(create_input_result.errors) > 0:
            raise ValueError(f"Could not create agent_element_input: {create_input_result.errors}")

    class CreatePipelineElementOutputPayload(GQLBaseModel):
        errors: List[str]

    for i, output in enumerate(outputs):
        create_output_result = client.createPipelineElementOutput(
            return_type=CreatePipelineElementOutputPayload,
            machineAgentVersionId=machine_agent_version_id,
            name=output,
            arrayNestingLevels=0,
            sortOrder=str(i),
            isAnnotation=True,
        )
        if len(create_output_result.errors) > 0:
            raise ValueError(f"Could not create agent_element_output: {create_output_result.errors}")

    logger.info(f"Machine Agent Version {machine_agent_version_id} created and published successfully.")
    endpoint = urlparse(client.endpoint_url)
    endpoint_domain = f"{endpoint.scheme}://{endpoint.netloc}"
    logger.info(
        "You can inspect this machine agent version at "
        f"{endpoint_domain}/admin/machine_agent_versions/{machine_agent_version_id}"
    )

    return machine_agent_version_id



"""Unified context builder with frame windowing, taxonomy, and Jinja2 support."""

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Set
from uuid import UUID

import celpy

from highlighter.agent.observations_table import ObservationsTable
from highlighter.client.base_models.entities import Entities
from highlighter.core.data_models.data_sample import DataSample

__all__ = [
    "ContextBuilder",
    "TaxonomyConfig",
    "ContentFilterConfig",
    "TemplateContext",
    "Frame",
]


@dataclass
class TaxonomyConfig:
    """Taxonomy configuration for context rendering"""

    # Object classes with descriptions
    object_classes: Dict[str, str] = field(default_factory=dict)
    # Attributes with descriptions
    attributes: Dict[str, str] = field(default_factory=dict)
    # Input taxonomy: maps object classes to their input attributes
    input_taxonomy: Dict[str, List[str]] = field(default_factory=dict)
    # Output taxonomy: defines expected output structure
    output_taxonomy: Dict[str, Dict[str, Any]] = field(default_factory=dict)


@dataclass
class ContentFilterConfig:
    """Configuration for filtering one specific content type to manage memory

    Each ContentFilterConfig governs buffering for ONE content type (video, audio,
    observations, text). Use multiple configs for different content types flowing
    through the pipeline.

    This prevents memory overflow by controlling what gets added to ContextBuilder.

    Examples:
        # Video stream buffering - HD only, every 5th frame, max 20
        ContentFilterConfig(
            content_type="video",
            filter_expr="width >= 1280 && height >= 720",
            sample_rate=5,
            max_samples=20
        )

        # Image buffering - high-res only, every image, max 10
        ContentFilterConfig(
            content_type="image",
            filter_expr="width >= 1920 && height >= 1080",
            sample_rate=1,
            max_samples=10
        )

        # Audio buffering - every 10th sample, max 10
        ContentFilterConfig(
            content_type="audio",
            sample_rate=10,
            max_samples=10
        )

        # Observations - keep last 50
        ContentFilterConfig(
            content_type="observations",
            max_samples=50
        )

        # Text - include all, no limits
        ContentFilterConfig(
            content_type="text"
        )

    Available CEL fields from DataSample:
        - content_type: str (e.g., "image", "video", "text", "audio")
        - stream_frame_index: int
        - media_frame_index: int
        - recorded_at: timestamp
        - width: int (for image/video)
        - height: int (for image/video)
    """

    # Required: Which content type this filter applies to
    content_type: str

    # Optional CEL expression for advanced filtering
    filter_expr: Optional[str] = None

    # Optional: Sample rate (1 = include all, 5 = every 5th, etc.)
    sample_rate: int = 1

    # Optional: Maximum samples to buffer (if not specified, no limit)
    max_samples: Optional[int] = None

    def __post_init__(self):
        """Initialize counters and compile CEL expression"""
        self._sample_counter: int = 0
        self._total_counter: int = 0

        # Compile CEL expression if provided
        self._cel_program = None
        if self.filter_expr:
            try:
                env = celpy.Environment()
                ast = env.compile(self.filter_expr)
                self._cel_program = env.program(ast)
            except Exception as e:
                raise ValueError(f"Invalid CEL filter expression for '{self.content_type}': {e}")

    def _to_cel_context(self, data_sample: DataSample) -> dict:
        """Convert DataSample to CEL context dict

        Args:
            data_sample: DataSample to convert

        Returns:
            Dictionary for CEL evaluation
        """
        ctx = {
            "content_type": data_sample.content_type,
            "stream_frame_index": data_sample.stream_frame_index,
            "media_frame_index": data_sample.media_frame_index,
            "recorded_at": data_sample.recorded_at,
        }

        # Add width/height for image/video content
        if data_sample.content_type.startswith(("image", "video")) and data_sample.wh:
            ctx["width"] = data_sample.wh[0]
            ctx["height"] = data_sample.wh[1]

        return celpy.json_to_cel(ctx)

    def should_include(self, data_sample: DataSample) -> bool:
        """Determine if a data sample should be included based on filter rules

        Evaluation order:
        1. CEL expression (if provided) - must pass
        2. max_samples limit - must not exceed
        3. sample_rate sampling - determines final inclusion

        Args:
            data_sample: DataSample to check

        Returns:
            True if sample should be included, False otherwise
        """
        # 1. Evaluate CEL expression if provided
        if self._cel_program:
            try:
                ctx = self._to_cel_context(data_sample)
                matches = self._cel_program.evaluate(ctx)
                if not matches:
                    return False
            except (celpy.evaluation.CELEvalError, TypeError, KeyError):
                # If CEL evaluation fails, reject the sample
                return False

        # 2. Check max_samples limit
        if self.max_samples is not None and self._total_counter >= self.max_samples:
            return False

        # 3. Apply sample rate
        self._sample_counter += 1

        if self._sample_counter >= self.sample_rate:
            self._sample_counter = 0
            self._total_counter += 1
            return True

        return False

    def reset(self):
        """Reset counters for next window"""
        self._sample_counter = 0
        self._total_counter = 0


class ContextBuilder:
    """Unified context builder with frame windowing and Jinja2 template support

    Manages memory by buffering different content types (video, audio, observations, text)
    with independent filtering strategies to prevent overflow.

    Features:
    - Frame windowing: Accumulate N frames before processing
    - Time windowing: Accumulate frames over T seconds before processing
    - Per-type buffering: Each content type has its own filter configuration
    - CEL expressions: Advanced filtering per content type
    - Sample rate control: Subsample high-frequency streams
    - Max samples: Cap buffer size per content type
    - Taxonomy support: Domain knowledge for complex use cases
    - Jinja2 templates: Flexible user-defined prompt formatting
    - Observation access: Clean API for entity observations

    Usage:
        # Buffer by frame count with content filters
        builder = ContextBuilder(
            frame_window_size=10,
            content_filters=[
                # Video: HD only, every 5th frame, max 20
                ContentFilterConfig(
                    content_type="video",
                    filter_expr="width >= 1280 && height >= 720",
                    sample_rate=5,
                    max_samples=20
                ),
                # Images: High-res only, every image, max 10
                ContentFilterConfig(
                    content_type="image",
                    filter_expr="width >= 1920 && height >= 1080",
                    sample_rate=1,
                    max_samples=10
                ),
                # Audio: Every 10th sample, max 10
                ContentFilterConfig(
                    content_type="audio",
                    sample_rate=10,
                    max_samples=10
                ),
                # Observations: Keep last 50
                ContentFilterConfig(
                    content_type="observations",
                    max_samples=50
                ),
                # Text: Include all, no limits
                ContentFilterConfig(
                    content_type="text"
                )
            ],
            taxonomy=TaxonomyConfig(...),
            stream_id="stream_123"
        )

        # Buffer by time (5 seconds)
        builder = ContextBuilder(
            time_window_seconds=5.0,
            content_filters=[
                # Video: HD only, every 5th frame, max 20
                ContentFilterConfig(
                    content_type="video",
                    filter_expr="width >= 1280 && height >= 720",
                    sample_rate=5,
                    max_samples=20
                ),
                # Images: High-res only, every image, max 10
                ContentFilterConfig(
                    content_type="image",
                    filter_expr="width >= 1920 && height >= 1080",
                    sample_rate=1,
                    max_samples=10
                ),
                # Audio: Every 10th sample, max 10
                ContentFilterConfig(
                    content_type="audio",
                    sample_rate=10,
                    max_samples=10
                ),
                # Observations: Keep last 50
                ContentFilterConfig(
                    content_type="observations",
                    max_samples=50
                ),
                # Text: Include all, no limits
                ContentFilterConfig(
                    content_type="text"
                )
            ],
            taxonomy=TaxonomyConfig(...),
            stream_id="stream_123"
        )

        # Buffer by both frame and time (whichever comes first)
        builder = ContextBuilder(
            frame_window_size=100,
            time_window_seconds=5.0,
            content_filters=[
                # Video: HD only, every 5th frame, max 20
                ContentFilterConfig(
                    content_type="video",
                    filter_expr="width >= 1280 && height >= 720",
                    sample_rate=5,
                    max_samples=20
                ),
                # Images: High-res only, every image, max 10
                ContentFilterConfig(
                    content_type="image",
                    filter_expr="width >= 1920 && height >= 1080",
                    sample_rate=1,
                    max_samples=10
                ),
                # Audio: Every 10th sample, max 10
                ContentFilterConfig(
                    content_type="audio",
                    sample_rate=10,
                    max_samples=10
                ),
                # Observations: Keep last 50
                ContentFilterConfig(
                    content_type="observations",
                    max_samples=50
                ),
                # Text: Include all, no limits
                ContentFilterConfig(
                    content_type="text"
                )
            ],
            taxonomy=TaxonomyConfig(...),
            stream_id="stream_123"
        )

        # Accumulate frames (filtering applied automatically per content type)
        builder.add_to_context(data_samples, entities)
        builder.add_to_context(data_samples, entities)

        # Build context when window is full
        if builder.is_window_full():
            context = builder.build_context()
            # Use with TemplateManager:
            # rendered = template_manager.render("prompt", context)
            builder.clear()  # Resets all filter counters
    """

    def __init__(
        self,
        frame_window_size: Optional[int] = None,
        time_window_seconds: Optional[float] = None,
        taxonomy: Optional[TaxonomyConfig] = None,
        content_filters: Optional[List[ContentFilterConfig]] = None,
        stream_id: Optional[str] = None,
    ):
        """Initialize ContextBuilder

        Args:
            frame_window_size: Number of frames to accumulate before rendering (default: 1)
            time_window_seconds: Number of seconds to accumulate before rendering (optional)
            taxonomy: Optional taxonomy configuration for domain knowledge
            content_filters: Optional list of per-content-type filter configurations
            stream_id: Optional stream identifier

        Note:
            - If both frame_window_size and time_window_seconds are specified, the window
              is considered full when EITHER condition is met (whichever comes first)
            - If neither is specified, defaults to frame_window_size=1
        """
        # Default to frame_window_size=1 if neither is specified
        if frame_window_size is None and time_window_seconds is None:
            frame_window_size = 1

        self.frame_window_size = frame_window_size
        self.time_window_seconds = time_window_seconds
        self.taxonomy = taxonomy or TaxonomyConfig()
        self.stream_id = stream_id

        # Build lookup dictionary by content_type for fast filtering
        self._filters_by_type: Dict[str, ContentFilterConfig] = {}
        if content_filters:
            for filter_config in content_filters:
                self._filters_by_type[filter_config.content_type] = filter_config

        # Accumulated data
        self._data_samples: List[DataSample] = []
        self._entities_buffer: Entities = Entities()
        self._template_vars: Dict[str, Any] = {}
        self._window_start_time: Optional[datetime] = None

    def add_to_context(
        self,
        data_samples: List[DataSample],
        entities: Optional[Entities] = None,
        template_vars: Optional[Dict[str, Any]] = None,
    ):
        """Add data samples and entities to the context window

        Data samples are filtered according to their content_type filter configuration.
        If no filter exists for a content_type, all samples of that type are included.

        Args:
            data_samples: List of DataSample objects
            entities: Optional dict of entities by ID
            template_vars: Optional custom template variables
        """
        # Apply per-content-type filtering
        filtered_samples = []
        for sample in data_samples:
            # Look up filter for this content type
            content_filter = self._filters_by_type.get(sample.content_type)

            if content_filter:
                # Apply filter if one exists for this content type
                if content_filter.should_include(sample):
                    filtered_samples.append(sample)
            else:
                # No filter for this content type - include by default
                filtered_samples.append(sample)

        # Set window start time on first sample if using time window
        # For multi-stream aggregation, use the earliest timestamp from the first batch
        if self.time_window_seconds is not None and self._window_start_time is None and filtered_samples:
            # Initialize with the earliest timestamp from this first batch
            self._window_start_time = min(sample.recorded_at for sample in filtered_samples)

        self._data_samples.extend(filtered_samples)

        if entities:
            self._entities_buffer.merge(entities, strategy="append")

        if template_vars:
            self._template_vars.update(template_vars)

    def is_window_full(self, current_sample_time: Optional[datetime] = None) -> bool:
        """Check if frame or time window is full

        Args:
            current_sample_time: Optional timestamp of current sample being processed.
                                 Used to check time window even if sample hasn't been added yet.

        Returns:
            True if:
            - frame_window_size is set and accumulated frames >= frame_window_size, OR
            - time_window_seconds is set and time elapsed >= time_window_seconds
            When both are set, returns True if EITHER condition is met.
        """
        # Check frame count condition
        if self.frame_window_size is not None:
            if len(self._data_samples) >= self.frame_window_size:
                return True

        # Check time window condition
        if self.time_window_seconds is not None and self._window_start_time is not None:
            # Use current_sample_time if provided, otherwise find the latest sample in buffer
            if current_sample_time:
                elapsed = (current_sample_time - self._window_start_time).total_seconds()
                if elapsed >= self.time_window_seconds:
                    return True
            elif self._data_samples:
                # Use the latest timestamp across all accumulated samples
                # This handles multi-stream aggregation where samples may arrive out of order
                latest_time = max(sample.recorded_at for sample in self._data_samples)
                elapsed = (latest_time - self._window_start_time).total_seconds()
                if elapsed >= self.time_window_seconds:
                    return True

        return False

    def build_context(self) -> "TemplateContext":
        """Build template context from accumulated data

        Returns:
            TemplateContext object for use with Jinja2 templates
        """

        return TemplateContext(
            data_samples=self._data_samples,
            entities=self._entities_buffer,
            stream_id=self.stream_id or "unknown",
            taxonomy=self.taxonomy,
            custom_vars=self._template_vars,
        )

    def clear(self):
        """Clear accumulated data and reset all content filter counters"""
        self._data_samples.clear()
        self._entities_buffer.clear()
        self._window_start_time = None
        # Keep template_vars and taxonomy (they're configuration)
        # Reset all content filter counters for next window
        for content_filter in self._filters_by_type.values():
            content_filter.reset()

    def get_accumulated_count(self) -> int:
        """Get number of accumulated data samples"""
        return len(self._data_samples)

    def get_data_samples(self) -> List[DataSample]:
        """Get accumulated data samples

        Returns:
            Direct reference to accumulated DataSample list (not a copy)
        """
        return self._data_samples


class TemplateContext:
    """Template context for Jinja2 rendering with full feature support

    Provides access to:
    - Frames (data samples with observations)
    - Taxonomy (object classes, attributes, input/output schemas)
    - Custom variables
    - Observation tables with CEL filtering

    Example template:
        # System Prompt
        You are analyzing {{ context.taxonomy.object_classes | length }} object types.

        # Taxonomy
        {% for class_name, description in context.taxonomy.object_classes.items() %}
        - {{ class_name }}: {{ description }}
        {% endfor %}

        # Frames
        {% for frame in context.frames %}
        Frame {{ frame.index }}: {{ frame.observations | length }} observations
        {% endfor %}

        # High-confidence detections using CEL filtering
        {% set high_conf = context.observations.filter("attribute.object_class.confidence >= 0.8") %}
        {% for obs in high_conf %}
        - {{ obs.attribute.object_class.value.name }}
        {% endfor %}
    """

    def __init__(
        self,
        data_samples: List[DataSample],
        entities: Optional[Entities],
        stream_id: str,
        taxonomy: TaxonomyConfig,
        custom_vars: Dict[str, Any],
    ):
        self._data_samples = data_samples
        self._entities = entities
        self._stream_id = stream_id
        self._taxonomy = taxonomy
        self._custom_vars = custom_vars

        # Lazy-loaded properties
        self._frames = None
        self._observations = None

    @property
    def stream_id(self) -> str:
        """Current stream identifier"""
        return self._stream_id

    @property
    def taxonomy(self) -> TaxonomyConfig:
        """Taxonomy configuration

        Example:
            {% for class_name, desc in context.taxonomy.object_classes.items() %}
            - {{ class_name }}: {{ desc }}
            {% endfor %}
        """
        return self._taxonomy

    @property
    def frames(self) -> List["Frame"]:
        """All frames in the accumulated window

        Example:
            {% for frame in context.frames %}
            Frame {{ frame.index }}: {{ frame.width }}x{{ frame.height }}
            Observations: {{ frame.observations | length }}
            {% endfor %}
        """
        if self._frames is None:
            self._frames = [Frame(ds, self._get_frame_observations(ds)) for ds in self._data_samples]
        return self._frames

    @property
    def observations(self) -> ObservationsTable:
        """All observations across all frames

        Example:
            {% set high_conf = context.observations.filter("attribute.object_class.confidence >= 0.8") %}
            Found {{ high_conf | length }} high-confidence detections

            {% for obs in context.observations.filter("attribute.object_class.value.name == 'person'") %}
            - Person at ({{ obs.annotation.location.xmin }}, {{ obs.annotation.location.ymin }})
            {% endfor %}
        """
        if self._observations is None:
            # Create observations table directly from entities buffer to avoid duplication
            # In multi-stream mode, entities are accumulated across streams/frames, and creating
            # per-frame observation tables would duplicate the same entities multiple times.
            if not self._entities:
                self._observations = ObservationsTable(rows={})
            else:
                # Use the first data sample as reference for data_sample fields in observation rows
                # The actual observation timestamps come from annotation.occurred_at
                reference_sample = self._data_samples[0] if self._data_samples else None
                if reference_sample:
                    self._observations = self._entities.to_observations_table(
                        self._stream_id, reference_sample
                    )
                else:
                    self._observations = ObservationsTable(rows={})
        return self._observations

    @property
    def entities(self) -> Optional[Entities]:
        """Raw entities dictionary

        Provides direct access to all entities aggregated across streams.

        Example:
            {% if context.entities %}
            Total entities: {{ context.entities | length }}
            {% for entity_id, entity in context.entities.items() %}
            Entity {{ entity_id }}:
              - Annotations: {{ entity.annotations | length }}
              - Observations: {% for ann in entity.annotations %}{{ ann.observations | length }}{% endfor %}
            {% endfor %}
            {% endif %}
        """
        return self._entities

    @property
    def observation_rows(self) -> List:
        """All observation rows as a list (for easier template iteration)

        Use this when you want to iterate over all observations without CEL filtering.

        Example:
            {% for obs in context.observation_rows %}
            - {{ obs.attribute_id }}: {{ obs.value }}
            {% endfor %}
        """
        if self._observations is None:
            # Build observations table first
            _ = self.observations
        return list(self._observations._rows.values())

    def __getitem__(self, key: str):
        """Access custom template variables

        Example:
            Task: {{ context['task_description'] }}
        """
        return self._custom_vars.get(key)

    def _get_frame_observations(self, data_sample):
        """Get observations for specific frame"""
        if not self._entities:
            # Return empty ObservationsTable
            return ObservationsTable(rows={})

        obs_table = self._entities.to_observations_table(self._stream_id, data_sample)
        return obs_table


@dataclass
class Frame:
    """Single data sample with observations"""

    _data_sample: DataSample
    _observations_table: ObservationsTable

    @property
    def index(self) -> int:
        """Frame index in stream"""
        return self._data_sample.stream_frame_index

    @property
    def type(self) -> str:
        """Content type (e.g., 'image', 'video', 'text')"""
        return self._data_sample.content_type

    @property
    def timestamp(self):
        """Frame timestamp"""
        return self._data_sample.recorded_at

    @property
    def width(self) -> Optional[int]:
        """Frame width (for image/video)"""
        if self.type.startswith(("image", "video")):
            return self._data_sample.wh[0]
        return None

    @property
    def height(self) -> Optional[int]:
        """Frame height (for image/video)"""
        if self.type.startswith(("image", "video")):
            return self._data_sample.wh[1]
        return None

    @property
    def content(self):
        """Frame content (use with caution in templates)"""
        return self._data_sample.content

    @property
    def observations(self) -> ObservationsTable:
        """Observations for this frame"""
        return self._observations_table



import logging
from abc import abstractmethod
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from highlighter.client.evaluation import EvaluationMetricResult
from highlighter.client.gql_client import HLClient
from highlighter.client.training_config import TrainingConfigType
from highlighter.core.labeled_uuid import LabeledUUID
from highlighter.datasets.dataset import Dataset


class BaseTrainer:

    def __init__(
        self,
        training_run_dir: Path,
        highlighter_training_config: TrainingConfigType,
    ):
        self._logger = logging.getLogger(__name__)
        self._training_run_dir = training_run_dir
        self._hl_training_config = highlighter_training_config

    @property
    def _hl_cache_dir(self):
        d = self._training_run_dir / ".hl"
        d.mkdir(exist_ok=True, parents=True)
        return d

    @property
    def _hl_cache_dataset_dir(self):
        d = self._hl_cache_dir / "datasets"
        d.mkdir(exist_ok=True, parents=True)
        return d

    @staticmethod
    def hl_training_config_path(training_run_dir: Path):
        d = training_run_dir / "training_config.yaml"
        return d

    def generate_boilerplate(self):
        """Puts the standard boilerplate files and makes directories required
        by in the Trainer in the training_run_dir. Downloads as caches
        the dataset annotations
        """
        self._hl_training_config.dump("yaml", BaseTrainer.hl_training_config_path(self._training_run_dir))

    @property
    @abstractmethod
    def training_data_dir(self) -> Path:
        """Path to data used to store training data"""
        pass

    @abstractmethod
    def generate_trainer_specific_dataset(self, hl_dataset: Dataset):
        pass

    @abstractmethod
    def train(self) -> Any:
        """Trains the Model, return the trained model instance"""
        pass

    @abstractmethod
    def evaluate(
        self, checkpoint: Path, cfg_path: Optional[Path] = None
    ) -> Dict[str, EvaluationMetricResult]:
        """Evaluate a model given a Path to a checkpoint, onnx_file or model instance"""
        pass

    @abstractmethod
    def export_to_onnx(self, trained_model) -> Path:
        """Export the trained model to onnx

        Return:
            Path to the onnx model
        """
        pass

    @abstractmethod
    def make_artefact(self, onnx_file_path: Path) -> Path:
        """Create a Highlighter training run artefact give the input checkpoint

        Return:
            Path to the artefact.yaml
        """
        pass

    @property
    def training_run_id(self) -> int:
        return self._hl_training_config.training_run_id

    @property
    def research_plan_id(self) -> int:
        return self._hl_training_config.evaluation_id

    def get_datasets(self, client: HLClient, page_size=20) -> Dataset:
        """Returns a Highlighter SDK Dataset object containing data from each
        split in the one object. `dataset.data_files_id.split` identifies the
        which split each data_file belongs to.
        """
        datasets = Dataset.read_training_config(
            client, self._hl_training_config, self._hl_cache_dataset_dir, page_size=page_size
        )
        return self._combine_hl_datasets(datasets)

    def _combine_hl_datasets(self, datasets):
        # When creating a training run in Highlighter the train split is required
        # but a user can supply either a test or dev set, or both. If not both we
        # duplicate the one that exists here
        if "test" not in datasets:
            datasets["test"] = deepcopy(datasets["dev"])
            datasets["test"].data_files_df.split = "test"
        if "dev" not in datasets:
            datasets["dev"] = deepcopy(datasets["test"])
            datasets["dev"].data_files_df.split = "dev"

        combined_ds = datasets["train"]
        combined_ds.append([datasets["dev"], datasets["test"]])

        return combined_ds

    def filter_dataset(self, dataset: Dataset) -> Dataset:
        """Optionally add some code to filter the Highlighter Datasets as required.
        The YoloWriter will only use entities with both a pixel_location attribute
        and a 'category_attribute_id' attribute when converting to the Yolo dataset format.
        It will the unique values for the object_class attribute as the detection
        categories.

        For example, if you want to train a detector that finds Apples and Bananas,
        and your taxonomy looks like this:

            - object_class: Apple
            - object_class: Orange
            - object_class: Banana

        Then you may do something like this:

            adf = combined_ds.annotations_df
            ddf = combined_ds.data_files_df

            orange_entity_ids = adf[(adf.attribute_id == OBJECT_CLASS_ATTRIBUTE_UUID) &
                                   (adf.value == "Orange")].entity_id.unique()

            # Filter out offending entities
            adf = adf[adf.entity_id.isin(orange_entity_ids)]

            # clean up images that are no longer needed
            ddf = ddf[ddf.data_file_id.isin(adf.data_file_id)]

            combined_ds.annotations_df = adf
        """

        # keep entities with specified categories
        categories = self.get_categories()

        def keep_entities_with_categories(e):
            for attr_id, value in categories:
                if e.attributes.get(attr_id, None) == value:
                    return True
            return False

        dataset.filter_entities(keep_entities_with_categories)

        return dataset

    def _train(self, hl_dataset: Dataset):

        self.generate_trainer_specific_dataset(hl_dataset)

        trained_model = self.train()
        onnx_model_path = self.export_to_onnx(trained_model)
        artefact_path = self.make_artefact(onnx_model_path)
        eval_metric_results = self.evaluate(onnx_model_path)
        return trained_model, artefact_path.absolute(), onnx_model_path, eval_metric_results

    def get_categories(self) -> List[Tuple[LabeledUUID, LabeledUUID]]:
        """Return a list of attribute_id, value tuples representing the
        unique categories for training
        """
        head_outputs = self._hl_training_config.input_output_schema.get_head_model_outputs(
            self._hl_training_config.head_idx
        )
        return [
            (
                LabeledUUID(o.taxon[-1].entity_attribute_id, label=o.taxon[-1].entity_attribute_name),
                LabeledUUID(
                    o.taxon[-1].entity_attribute_enum_id, label=o.taxon[-1].entity_attribute_enum_value
                ),
            )
            for o in head_outputs
        ]



import json
import logging
import tempfile
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml
from pydantic import ConfigDict, Field, field_validator

from ..core import DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE, GQLBaseModel
from .aws_s3 import upload_file_to_s3
from .gql_client import HLClient
from .io import download_bytes

logger = logging.getLogger(__name__)

__all__ = [
    "TrainingRunArtefactTypeEnum",
    "TrainingRunArtefactType",
    "TrainingRunType",
    "get_training_run_artefacts",
    "get_training_run_artefact",
    "get_latest_training_run_artefact",
    "get_capability_implementation_file_url",
    "create_training_run_artefact",
]


def _get_now_str():
    return datetime.now().isoformat()


class TrainingRunArtefactTypeEnum(str, Enum):
    OnnxOpset11 = "OnnxOpset11"
    OnnxOpset14 = "OnnxOpset14"
    TorchScriptV1 = "TorchScriptV1"
    TensorFlowV1 = "TensorFlowV1"
    DeprecatedMmpond = "DeprecatedMmpond"
    DeprecatedClassilvier = "DeprecatedClassilvier"
    DeprecatedSilverclassify = "DeprecatedSilverclassify"
    OnnxRuntimeAmd64 = "OnnxRuntimeAmd64"
    OnnxRuntimeArm = "OnnxRuntimeArm"
    DeprecatedCapabilityImplementationFile = DEPRECATED_CAPABILITY_IMPLEMENTATION_FILE
    TextEmbedder = "TextEmbedder"


class TrainingRunArtefactType(GQLBaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: Optional[str] = None
    file_url: str
    type: TrainingRunArtefactTypeEnum
    updated_at: Optional[str] = Field(default_factory=_get_now_str)
    inference_config: Dict = {}  # Graphql expects at least an empty dict not None
    training_config: Dict = {}  # Graphql expects at least an empty dict not None
    supporting_files: Dict = {}  # Graphql expects at least an empty dict not None

    @field_validator("inference_config", "training_config", "supporting_files", mode="before")
    @classmethod
    def none_to_empty_dict(cls, v):
        if v is None:
            return {}
        return v

    @classmethod
    def from_yaml(cls, path: Union[Path, str]):
        path = Path(path)
        with path.open("r") as f:
            data = yaml.safe_load(f)
        return cls(**data)

    def dump_yaml(self, path: Union[Path, str]):
        path = Path(path)
        with path.open("w") as f:
            yaml.dump(self.model_dump(), f)

    def create(self, client: HLClient, training_run_id: int):
        """Create and upload a training run artefact to Highlighter.

        This method handles the complete workflow of artefact creation including:
        - Uploading the artefact file to S3
        - Creating the artefact record in Highlighter via GraphQL
        - Setting up the associated capability for inference

        Args:
            client: Authenticated HLClient instance for GraphQL operations
            training_run_id: ID of the training run to associate this artefact with

        Returns:
            The created training run artefact with assigned ID

        Raises:
            Exception: If artefact creation fails or returns errors
        """
        from .capabilities import create_capability_for_artefact

        artefact_file_data = upload_file_to_s3(
            client,
            self.file_url,
            mimetype="application/octet-stream",
        )
        logger.info(f"Created file in s3:\n{artefact_file_data}")

        class _TrainingRunArtefactType(GQLBaseModel):
            id: str

        class CreateTrainingRunPayload(GQLBaseModel):
            errors: List[str]
            training_run_artefact: Optional[_TrainingRunArtefactType] = None

        artefact_result = client.createTrainingRunArtefact(
            return_type=CreateTrainingRunPayload,
            trainingRunId=training_run_id,
            type=self.type,
            fileData=artefact_file_data,
            inferenceConfig=self.inference_config,
            trainingConfig=self.training_config,
        )
        logger.info(f"create artefact result: {artefact_result}")

        create_capability_for_artefact(
            client=client,
            artefact_id=artefact_result.training_run_artefact.id,
            training_run_id=training_run_id,
            inference_config=self.inference_config,
        )
        return artefact_result.training_run_artefact


class TrainingRunType(GQLBaseModel):
    model_config = ConfigDict(use_enum_values=True, protected_namespaces=())

    id: int
    name: str
    updated_at: str
    model_implementation_file_url: Optional[str] = None
    training_run_artefacts: List[TrainingRunArtefactType] = []
    training_config: Optional[Any] = None


def _download_training_run_artefact(
    training_run_artefact: TrainingRunArtefactType,
    file_url_save_path: Optional[str] = None,
):
    if file_url_save_path is None:
        file_url_save_path = f"{tempfile.mkdtemp()}/{training_run_artefact.id}"

    download_bytes(training_run_artefact.file_url, Path(file_url_save_path))

    training_run_artefact.file_url = str(file_url_save_path)
    return training_run_artefact


def get_training_run_artefacts(
    hl_client: HLClient,
    training_run_id: int,
    filter_by_artefact_type: Optional[Union[str, TrainingRunArtefactTypeEnum]] = None,
) -> List[TrainingRunArtefactType]:
    """Return a list of artefacts for a training run
    sorted newest to oldest according to TrainingRunArtefactType.updated_at

    If `filter_by_artefact_type` is set then will only retrun artefacts with
    matching TrainingRunArtefactType.type
    """

    training_run: TrainingRunType = hl_client.trainingRun(
        return_type=TrainingRunType,
        id=training_run_id,
    )

    if training_run.model_implementation_file_url is not None:
        # If TrainingRun.updated_at is None then we set it to
        # datetime.min for the purposes of sorting.
        capability_implementation_file_artefact: TrainingRunArtefactType
        capability_implementation_file_artefact = TrainingRunArtefactType(
            id=None,
            file_url=training_run.model_implementation_file_url,
            type=TrainingRunArtefactTypeEnum.DeprecatedCapabilityImplementationFile,
            updated_at=datetime.min.isoformat() if training_run.updated_at else training_run.updated_at,
            inference_config={},
            training_config={},
            supporting_files={},
        )
        training_run.training_run_artefacts.append(capability_implementation_file_artefact)

    if filter_by_artefact_type is not None:
        artefact_type = TrainingRunArtefactTypeEnum(filter_by_artefact_type)
        training_run.training_run_artefacts = [
            a for a in training_run.training_run_artefacts if a.type == artefact_type
        ]

    return sorted(training_run.training_run_artefacts, key=lambda x: x.updated_at, reverse=True)


def get_training_run_artefact(
    hl_client: HLClient,
    training_run_artefact_id: str,
    download_file_url: bool = False,
    file_url_save_path: Optional[str] = None,
) -> TrainingRunArtefactType:
    """Get the TrainingRunArtefact object for a given training_run_artefact_id.

    If `download_file_url` is `True` then download either to a temporary
    directory or `file_url_save_path`. Once downloaded TrainingRunArtefact.file_url
    will be updated to the location where is has been downloaded to.
    """

    training_run_artefact: TrainingRunArtefactType
    training_run_artefact = hl_client.trainingRunArtefact(
        return_type=TrainingRunArtefactType, id=training_run_artefact_id
    )

    if download_file_url:
        training_run_artefact = _download_training_run_artefact(
            training_run_artefact, file_url_save_path=file_url_save_path
        )

    return training_run_artefact


def get_latest_training_run_artefact(
    hl_client: HLClient,
    training_run_id: int,
    download_file_url: bool = False,
    file_url_save_path: Optional[str] = None,
    filter_by_artefact_type: Optional[Union[str, TrainingRunArtefactTypeEnum]] = None,
) -> Optional[TrainingRunArtefactType]:
    """Get the TrainingRunArtefact with the most recent updated_at value.

    If `download_file_url` is `True` then download either to a temporary
    directory or `file_url_save_path`. Once downloaded TrainingRunArtefact.file_url
    will be updated to the location where is has been downloaded to.
    """
    training_run_artefacts: List[TrainingRunArtefactType] = get_training_run_artefacts(
        hl_client, training_run_id, filter_by_artefact_type=filter_by_artefact_type
    )

    if len(training_run_artefacts) == 0:
        return None
    training_run_artefact: TrainingRunArtefactType = training_run_artefacts[0]

    if download_file_url:
        abs_file_url_save_path = str(Path(file_url_save_path).absolute())
        training_run_artefact = _download_training_run_artefact(
            training_run_artefact, file_url_save_path=abs_file_url_save_path
        )

    return training_run_artefact


def get_capability_implementation_file_url(
    hl_client: HLClient,
    training_run_id: int,
) -> Optional[str]:
    result: TrainingRunType = hl_client.trainingRun(
        return_type=TrainingRunType,
        id=training_run_id,
    )
    return result.model_implementation_file_url


def create_training_run_artefact(
    hl_client: HLClient,
    training_run_id: int,
    artefact_path: str,
    training_run_artefact_type: TrainingRunArtefactTypeEnum,
    checkpoint_name: str,
    training_config: Union[Path, str, Dict],
    inference_config: Union[Path, str, Dict],
    data_source_uuid: Optional[str] = None,
) -> TrainingRunArtefactType:
    if not isinstance(training_config, dict):
        training_config_path = Path(training_config)
        with training_config_path.open("r") as f:
            training_config = {
                training_config_path.name: f.read(),
            }

    if not isinstance(inference_config, dict):
        inference_config_path = Path(inference_config)
        with inference_config_path.open("r") as f:
            loaders = {
                ".yaml": yaml.safe_load,
                ".yml": yaml.safe_load,
                ".json": json.load,
            }

            inference_config = loaders[inference_config_path.suffix](f)

    artefact_file_data = upload_file_to_s3(
        hl_client,
        artefact_path,
        data_source_uuid=data_source_uuid,
    )
    print(f"Created file in s3:\n{artefact_file_data}")

    class CreateTrainingRunPayload(GQLBaseModel):
        errors: List[str]
        training_run_artefact: Optional[TrainingRunArtefactType] = None

    result = hl_client.createTrainingRunArtefact(
        return_type=CreateTrainingRunPayload,
        trainingRunId=training_run_id,
        type=training_run_artefact_type,
        checkpoint=checkpoint_name,
        fileData=artefact_file_data,
        inferenceConfig=inference_config,
        trainingConfig=training_config,
    )
    if result.errors:
        raise ValueError(f"{result.errors}")

    return result.training_run_artefact



from typing import (
    Any,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    TypeAlias,
    Union,
)

import numpy as np
from PIL import Image
from pydantic import BaseModel, Field, PrivateAttr, ValidationInfo, field_validator
from shapely import geometry

from highlighter.client.base_models.annotation import Annotation
from highlighter.client.base_models.datum_source import DatumSource
from highlighter.client.base_models.entity import Entity
from highlighter.core.data_models.data_sample import DataSample
from highlighter.core.labeled_uuid import LabeledUUID
from highlighter.datasets.cropping import CropArgs
from highlighter.predictors.prediction import Prediction

Zone: TypeAlias = Union[Tuple[int, int, int, int] | Tuple[float, float, float, float]]


class CropProcessor(BaseModel):
    crop_args: CropArgs = Field(..., default_factory=lambda: CropArgs())
    zone_entity_ids: Optional[List[LabeledUUID]] = None
    zones: Optional[List[Zone]] = None

    class ZoneSpec(NamedTuple):
        entity_id: LabeledUUID
        location: geometry.Polygon

    _zones: Optional[List["CropProcessor.ZoneSpec"]] = PrivateAttr(default=None)

    @field_validator("zones", mode="after")
    @classmethod
    def set_zone_annotations(cls, zones: Optional[List[Zone]], info: ValidationInfo) -> Optional[List[Zone]]:
        """Instantiate zone specs for each zone"""
        if zones is None:
            zone_specs = None
        else:
            zone_entity_ids = info.data.get("zone_entity_ids")
            if zone_entity_ids:
                assert len(zone_entity_ids) == len(zones), "Must provide an entity_id for each zone, or None"
            else:
                zone_entity_ids = [LabeledUUID.uuid4(label=f"zone_{i}") for i, _ in enumerate(zones)]

            def make_zone_spec(zone, zone_entity_id):
                x0, y0, x1, y1 = zone
                location = geometry.Polygon([(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)])
                return cls.ZoneSpec(entity_id=zone_entity_id, location=location)

            zone_specs = [make_zone_spec(z, zid) for z, zid in zip(zones, zone_entity_ids)]

        info.data["_zones"] = zone_specs
        return zones

    def crop_annotations(
        self,
        annotations: Iterable[Annotation] = None,
    ) -> List[Union[np.ndarray | Annotation]]:
        """Crop each the Annotation.location for each Annotation.data_samples"""
        return [a.crop_image(crop_args=self.crop_args) for a in annotations]

    def crop_data_samples(
        self,
        data_samples: List[DataSample],
    ) -> Tuple[List[np.ndarray | Image.Image], List[ZoneSpec]]:
        """Crop each DataSample in data_samples with the locations in _zones

        Return a tuple of 2 flat lists:

            result[0]: a list of the cropped images
            result[1]: a list of Annotation objects corrisponding to each crop
        """

        if not self._zones:
            raise ValueError("No zones provided")

        # Create a copy of each zone spec for each DataSample. This is the
        # same as if a model produced N annotations per image.
        zone_specs = [[item for item in self._zones] for _ in data_samples]

        crop_imgs = []
        crop_src_zones = []
        for data_sample, specs in zip(data_samples, zone_specs):
            crop_locations = [s.location for s in specs]
            crops = data_sample.crop_content(crop_locations, crop_args=self.crop_args)
            crop_imgs.extend(crops)
            crop_src_zones.extend(specs)

        return crop_imgs, crop_src_zones
